package nemi.apriori;

import java.util.TreeMap;
import java.util.Vector;



/**
 * Graph/motif implementation using encoding from article.
 */
public class AprioriGraph implements Comparable {

	String stringCode;
	public AprioriGraphLabels labels;
	int codeLength;
	int n;
	boolean directed;

	public AprioriGraph( boolean[][] adjacency, AprioriGraphLabels labels, boolean directed) {
		//aprioriGraph = apriori;
		this.n = adjacency.length;
		this.directed = directed;
		this.stringCode = createStringCode(adjacency);
		this.labels = labels;
		this.codeLength = stringCode.length();
		
	}

	public AprioriGraph(String stringCode, AprioriGraphLabels labels, boolean directed) {
		//aprioriGraph = apriori;
		this.n = labels.labels.length(); 
		this.labels = labels;
		this.stringCode = stringCode;
		this.directed = directed;
		this.codeLength = stringCode.length();
		// (int)Math.sqrt(codeLength*2.0)
											// + 1;
	}

	public Vector<Integer> getColMatchIx(char forEdge, char forLabel,
			int fromLabelIx) {

		// codeString indexes for the hits (might be empty)
		Vector<Integer> hits = new Vector<Integer>();

		// traversing all labels
		for (int lblIx = fromLabelIx; lblIx < labels.labels.length(); lblIx++) {
			// label must be less than or equal
			if (labels.labels.charAt(lblIx) > forLabel) {
				break;
			} else if (labels.labels.charAt(lblIx) == forLabel) {
				int codeIx = lblIx * (lblIx - 1) / 2;
				for (int i = codeIx; i < codeIx + fromLabelIx; i++) {
					if (forEdge == stringCode.charAt(i)) {
						hits.add(i);
					}
				}
			}
		}
		return hits;
	}
	
	public static void main(String[] args) {
		AprioriGraph g = new AprioriGraph("110100",new AprioriGraphLabels("aaaa"), false );
		AprioriGraph g2 = g.getCanonicalForm();
		System.out.println(g2);
	}

    /**
     * Generate all normal form codes and return the smallest as the canonical form.
     * @return This motif's canonical form motif.
     */
    public AprioriGraph getCanonicalForm() {

		// first level: no nodes are connected to itself
		Vector<String> initCodes = new Vector<String>();
		// using graphlabel class to keep track of the labels' ordering
		Vector<AprioriGraphLabels> initLabels = new Vector<AprioriGraphLabels>();

		// initiating matrices and labels for the characters
		for (int i = 0; i < n; i++) {
			if (labels.labels.charAt(0) == labels.labels.charAt(i)) {
				initCodes.add(""); // no code for first chars
				Vector<Integer> lblPtr = new Vector<Integer>();
				lblPtr.add(i);
				initLabels.add(new AprioriGraphLabels(String.valueOf(labels.labels
						.charAt(i)), lblPtr));
			} else if (i > 0)
				break;
		}

		Vector<AprioriGraph> returns = new Vector<AprioriGraph>();
		for (int i = 0; i < initLabels.size(); i++) {
			// starting recursion
			recurseNormalForm(initCodes
					.toArray(new String[initCodes.size()]), initLabels
					.toArray(new AprioriGraphLabels[initLabels.size()]), i,
					returns);
		}

		AprioriGraph smallest = returns.get(0);
		for (AprioriGraph normalform : returns) {
			if (normalform.compareTo(smallest) < 0)
				smallest = normalform;
			//out("\nnormalform of:\n" + this + "\n" + normalform);
		}
		return smallest;
	}

	public int exists(AprioriGraph g2_temp) {
		return 1;
	}

    /**
     * Merge and recurse motifs, generating small codes.
     * @param matrices
     * @param labels
     * @param minIx
     * @param returns
     * @return A list of merged motifs.
     */
    private Vector<AprioriGraph> recurseNormalForm(String[] matrices,
			AprioriGraphLabels[] labels, int minIx, Vector<AprioriGraph> returns) {
		// no more mergins possible, return result
		if (labels.length > 0
				&& labels[0].labels.length() == this.labels.labels.length()) {
			returns.add(new AprioriGraph(matrices[0], labels[0], this.directed));
			return returns;
		}
		// only one matrix left, add new edge(s if several equal) from g
		if (matrices.length == 1) {
			addNewEdges(matrices, labels, returns);
		}
		// more matrices to merge.. continue
		else {
			combineEdges(matrices, labels, minIx, returns);
		}
		return returns;
	}

	/**
	 * When creating the normal forms of an adj matrix, there are two
	 * phases. When no more permutations of equal characters are possible,
	 * the next character(s) must be added.
	 * 
	 * @param matrices
	 * @param newLabels
	 * @param returns
	 */
	private void addNewEdges(String[] matrices, AprioriGraphLabels[] newLabels,
			Vector<AprioriGraph> returns) {
		// current position in graph newLabels
		int gPos = newLabels[0].labels.length();
		// if there's several equal chars, recurse every single one of them
		int nSucc = 0;
		// counting the number of equal letters following each other at gPos
		for (int i = gPos; i < this.labels.labels.length()
				&& this.labels.labels.charAt(gPos) == this.labels.labels
						.charAt(i); i++) {
			nSucc++;
		}

		// the new label is equal for all new; only pointers differ
		String newLabel = newLabels[0].labels
				+ this.labels.labels.charAt(gPos);
		AprioriGraphLabels[] newGraphLabels = new AprioriGraphLabels[nSucc];
		for (int i = 0; i < nSucc; i++) {
			Vector<Integer> newPointers = new Vector<Integer>();
			// add all existing pointers
			newPointers.addAll(newLabels[0].graphPointers);
			// add pointer to new node
			newPointers.add(gPos + i);
			newGraphLabels[i] = new AprioriGraphLabels(newLabel, newPointers);
		}

		// adding last edge to matrices
		String[] newMatrices = new String[nSucc];
		for (int i = 0; i < newMatrices.length; i++) {
			int codeLength = newLabels[0].labels.length();
			int codeIx = (codeLength + i) * (codeLength + i - 1) / 2;
			String append = this.stringCode.substring(codeIx, codeIx
					+ codeLength);
			// appending new code to old
			newMatrices[i] = matrices[0] + append;
		}

		// recurse to get all combinations of the new letters
		for (int i = 0; i < newMatrices.length; i++) {
			recurseNormalForm(newMatrices, newGraphLabels, i, returns);
		}
	}

	/**
	 * A set of nodes are already added, and their permutations are
	 * calculated. The merging in figure 6 of the article.
	 * 
	 * @param matrices
	 * @param newLabels
	 * @param minIx
	 * @param returns
	 */
	private void combineEdges(String[] matrices, AprioriGraphLabels[] newLabels,
			int minIx, Vector<AprioriGraph> returns) {

		// new number of matrices
		int nRetMats = matrices.length - 1;
		// return matrices and newLabels
		String[] retMats = new String[nRetMats];
		AprioriGraphLabels[] retLbls = new AprioriGraphLabels[nRetMats];

		// remember the smallest matrixcode found
		TreeMap<Integer, String> ix2codeValue = new TreeMap<Integer, String>();
		String largestValue = "5";

		// previos smallest code's last nodelabel
		int smallerNodeNr = newLabels[minIx].lastNodePtr;

		// joining the code-retMin candidate with the others (figure 6)
		for (int retIx = 0, largerIx = 0; retIx < nRetMats; retIx++, largerIx++) {

			// jump over the previous minimum
			if (largerIx == minIx)
				largerIx++;

			// the last, larger node's code
			String largerCode = matrices[largerIx].substring(matrices[0]
					.length()
					- newLabels[0].labels.length() + 1, matrices[0]
					.length());

			// the last edges' value is missing, must be retrieved from g
			// (figure 6)
			int largerNodeNr = newLabels[largerIx].lastNodePtr;
			char lastNodeVal = getEdge(largerNodeNr, smallerNodeNr);

			// concatenating min, larger and last edge codes
			retMats[retIx] = matrices[minIx] + largerCode + lastNodeVal;

			// concat newLabels and corresponding pointers
			String newLabel = newLabels[minIx].labels
					+ newLabels[largerIx].labels
							.substring(newLabels[0].labels.length() - 1);
			Vector<Integer> newLabelPointers = new Vector<Integer>(
					newLabels[minIx].graphPointers);
			newLabelPointers.add(newLabels[largerIx].lastNodePtr);
			retLbls[retIx] = new AprioriGraphLabels(newLabel, newLabelPointers);

			// record code value and index; only the smallest are considered
			String thisVal = retMats[retIx];
			if (thisVal.compareTo(largestValue) < 0)
				largestValue = thisVal;
			ix2codeValue.put(retIx, thisVal);
		}

		// recurse all indexes with similar and smallest code values
		for (Integer i : ix2codeValue.keySet()) {
			String val = ix2codeValue.get(i);
			if (val.equals(largestValue)) {
				recurseNormalForm(retMats, retLbls, i, returns);
			}
		}
	}

	/**
	 * Gets the edge from a node to another node through the code
	 */
	public char getEdge(int from, int to) {
		if(from == to){
			System.out.println("ERROR: Checking diagonal from " +from + " to " + to );
			return '0';
		}
		//System.out.println("DEBUG: Checking  from " +from + " to " + to );
		
		boolean switched = false;
		if (from > to) {
			switched = true;
			int temp = from;
			from = to;
			to = temp;
		}
		int topRowIx = to * (to - 1) / 2; // pointing at (0, to) in matrix
		char ret = stringCode.charAt(topRowIx + from);
		// if directed edge and to/from switched, redirect arrow
		if (switched)
			if (ret == '3')
				ret = '2';
			else if (ret == '2')
				ret = '3';

		return ret;
	}


    /**
     * Generate an Apriori-code representation of the given matrix.
     * @param adjacency Matrix representation of the graph.
     * @return Apriori-representation.
     */
    private String createStringCode(boolean[][] adjacency) {
		// int retLenght = n * (n-1) / 2;
		// int codeix = 0;

		StringBuilder sb = new StringBuilder();
		
		if (directed) {
			for (int j = 1; j < n; j++) {
				for (int i = 0; i < j; i++) {
					if (adjacency[i][j]) {
						sb.append(adjacency[j][i] ? "3" : "1");
					} else {
						sb.append(adjacency[j][i] ? "2" : "0");
					}
				}
			}
		} else {
			for (int j = 1; j < n; j++) {
				System.out.println(j + "/" + n);
				for (int i = 0; i < j; i++) {
					sb.append(adjacency[i][j] ? "1" : "0");
				}
			}
		}
		
		return sb.toString();
	}

	public String toString() {
		return "Apriori.Graph instance: directed: " + directed
				+ ", stringCode: " + stringCode + ", labels: " + labels;
	}

    /**
     * Graphs having equal stingcodes are considered equal.
     * @param o
     * @return The string code comparison between this graph and the provided graph. 
     */
    public int compareTo(Object o) {
		if (o instanceof AprioriGraph) {
			AprioriGraph otherGraph = (AprioriGraph) o;
			// unequal lenghts not possible
			if (this.stringCode.length() != otherGraph.stringCode.length())
				return 0;

			return stringCode.compareTo(otherGraph.stringCode);
		}
		return 0; // To change body of implemented methods use File |
					// Settings | File Templates.
	}

}