package nemi.apriori;


import cern.colt.bitvector.BitVector;

import java.util.*;

/**
 * Apriori motif implementation, according to the article.
 */
public class GraphMotif extends SubGraphView {
    String labels;
    BitVector code;

    public GraphMotif(int nodes, String Labels) {
        code = new BitVector(nodes);
        setLabels(Labels);
    }

    private GraphMotif(BitVector v, String labels) {
        code = v;
        this.labels = labels;
    }

    public GraphMotif(String stringcode, String Labels) {
        code = new BitVector(stringcode.length());
        setLabels(Labels);
        if (!isCorrectLength()) {
            System.out.println("WRONG SIZE");
        }
        for (int i = 0; i < stringcode.length(); i++) {
            if (stringcode.charAt(i) == '1') {
                code.putQuick(i, true);
            }
        }
    }

    public String getLabels() {
        return labels;
    }

    /**
     * Join operation of two apriori-motifs. This motif has to be smaller than the given motif
     * to be able to join.
     * @param c2 The motif to join this motif with.
     * @return The join produce two new motifs, as the last edge is unknown.
     */
    public GraphMotif[] join(GraphMotif c2) {
        if (this.nodeCount() != c2.nodeCount()) {
            //System.out.println("Wrong nodecount");
            return null;
        }
        if (this.compareTo(c2) > 0) {
            //System.out.println("Label too low");
            return null;
        }
        if (!labels.substring(0, labels.length() - 1).equals(
                c2.labels.substring(0, labels.length() - 1))) {
            return null;
        }

        if (code.size() > 1) {
            long n1 = this.code.getLongFromTo(0, code.size() - this.nodeCount() - 1);
            long n2 = c2.code.getLongFromTo(0, code.size() - this.nodeCount() - 1);
            if (n1 != n2) return null;
        }

        String newlabels = new String(labels);
        newlabels += c2.labels.charAt(c2.labels.length() - 1);
        BitVector bv = code.copy();
        bv.setSize(code.size() + nodeCount());
        if (code.size() > 1) {
            bv.replaceFromToWith(code.size(), bv.size() - 2,
                    c2.code, c2.code.size() - c2.nodeCount() + 1);
        }
        //Vector<GraphMotif> returnVec = new Vector<GraphMotif>();
        bv.putQuick(bv.size() - 1, false);
        GraphMotif[] returnVec = new GraphMotif[2];
        returnVec[0] = new GraphMotif(bv.copy(), new String(newlabels));
        bv.putQuick(bv.size() - 1, true);
        returnVec[1] = new GraphMotif(bv, newlabels);

        return returnVec;
    }

    /**
     * Get the label of the given node-index.
     * @param index Node index.
     * @return The given node's label.
     */
    public String getLabelOf(int index) {
        return labels.substring(index, index + 1);
    }

    public void setLabels(String labels) {
        this.labels = labels;
    }


    public void setEdge(int i, int j, boolean value) {
        int index = indexOfEdgeBetween(i, j);
        code.putQuick(index, value);
    }

    public boolean getEdge(int i, int j) {
        int index = indexOfEdgeBetween(i, j);
        return code.getQuick(index);
    }

    public boolean isSorted() {
        char last = 0;
        for (int i = 0; i < labels.length(); i++) {
            char next = labels.charAt(i);
            if (last > next)
                return false;
            last = next;
        }
        return true;
    }

    /**
     * Verify that the motifs code length is valid.
     * @return Validity of the code length.
     */
    public boolean isCorrectLength() {
        int n = labels.length();
        int codelength = n * (n - 1) / 2;
        return code.size() == codelength;
    }

    public String getStringCode() {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < code.size(); i++) {
            sb.append(code.getQuick(i) ? "1" : "0");
        }
        return sb.toString();
    }

    public String toString() {
        StringBuilder sb = new StringBuilder(code.size() + labels.length());
        sb.append(labels);
        for (int i = 0; i < code.size(); i++) {
            sb.append(code.getQuick(i) ? "1" : "0");
        }
        return sb.toString();
    }

    public int compareTo(SubGraphView sg) {
        int var = this.nodeCount() - sg.nodeCount();
        if (var != 0) return var;
        return this.toString().compareTo(sg.toString());
    }

    @Override
    public Object clone() {
        // TODO Auto-generated method stub
        BitVector cl = (BitVector) code.clone();
        GraphMotif gm = new GraphMotif(0, "");
        gm.setLabels(new String(labels));
        gm.code = cl;
        return gm;

    }

    public static int indexOfEdgeBetween(int i, int j) {
        if (i > j) {
            int temp = i;
            i = j;
            j = temp;
        }
        return ((j * (j - 1)) / 2) + i;
    }

    public int nodeCount() {
        return labels.length();
    }

    public int hashCode() {
        int hash = code.hashCode();
        int labelcode = labels.hashCode();
        //System.out.println("hash" + hash*labelcode);
        return hash + labelcode;
    }

    public boolean equals(Object o) {
        if (o instanceof GraphMotif) {
            GraphMotif f = (GraphMotif) o;
            return this.code.equals(f.code) && this.labels.equals(f.getLabels());

        } else {
            return false;
        }
    }

    public void setSize(int i) {
        code.setSize(i);
    }
    public boolean isConnected() {
        if (labels.length() <= 1) return true;
        //HashSet<Node> nodes = new HashSet<Node>(subGraph);
        HashSet<Integer> seen = new HashSet<Integer>(labels.length());
        LinkedList<Integer> que = new LinkedList<Integer>();
        que.offer(0);
        seen.add(0);
        //Map<Node, Set<Node>> neighmap = Toolkit.getNodeNeighborMap(parentNet);
        while (que.size() > 0) {
            int node = que.poll();
            for(int i =0; i < labels.length();i++){
                if(seen.contains(i)) continue;
                if(getEdge(node,i)) {
                    seen.add(i);
                    que.offer(i);

                }
            }
        }
        return seen.size() == labels.length();
    }
}


