//$Id:IterativeClusterer.java 129 2006-02-21 16:33:10Z sundsdal $
package nemi.cluster;

import cytoscape.CyNetwork;
import giny.model.Node;
import nemi.NetNode;
import nemi.compare.CsaNeighComparer;

import java.util.*;


/**
 * Iterative clusterer using neighborhoods for scoring node similarities. When calculating similarity
 * between two nodes, our CSA maximum bipartite matching solver finds the best node to node
 * matching using the given comparer.
 */
public class CsaIterativeClusterer implements Clusterer {


    public CsaIterativeClusterer() {
    }

    /**
     * Cluster the given networks using node neighborhoods for similarity measure.
     * @param networks Networks to be clustered.
     * @param neighComp Single node to node comparer; used when calculating scores finding the optimal
     * bipartite match between the two nodes' neighborhoods.
     * @param compType The comparer type.
     * @param least Cluster limit. All nodes equal or more equal to least will be placed in the same
     * clusters.
     * @param joinSingleClusters When all nodes are clustered, some clusters may contain only one single
     * node. This flag joins all single-noded clusters.
     * @return A list of clusters. 
     */
    public List<Cluster> cluster(Collection<CyNetwork> networks, CsaNeighComparer neighComp,
                                 QualityAssesser.CompareClusterNodeType compType
            , double least, boolean joinSingleClusters) {

        Vector<NetNode> nodesNotInCluster = new Vector<NetNode>();

        //System.out.println("making networkset");
        HashSet<Cluster> clusters = new HashSet<Cluster>();
        int mycount = 0;
        for (CyNetwork network : networks) {
            nodesNotInCluster.ensureCapacity(network.getNodeCount() + nodesNotInCluster.size());
            for (Iterator<Node> nodeIter = network.nodesIterator(); nodeIter.hasNext();) {
                nodesNotInCluster.add(new NetNode(network, nodeIter.next()));
                mycount++;
            }
        }
        Vector<NetNode> nodesInCluster = new Vector<NetNode>(mycount);

        //System.out.println(least + " Nodecount " + mycount);
        //nodesInCluster.addAll(nodesNotInCluster);
        Vector<NetNode> nodesInClusterTemp = new Vector<NetNode>();
        while (nodesNotInCluster.size() > 0) {
            //System.out.println("rest:" + nodesNotInCluster.size());
            //System.out.flush();
            Iterator<NetNode> nIt = nodesNotInCluster.iterator();
            NetNode nStart = nIt.next();
            nIt.remove();
            Cluster c = new Cluster(nStart.node, nStart.net);
            nodesInClusterTemp.add(nStart);
            //try non-clustered nodes first
            for (Iterator<NetNode> notInClusterIter = nodesNotInCluster.iterator(); notInClusterIter.hasNext();)
            {
                NetNode n = notInClusterIter.next();
                double similarity =  neighComp.similarity(nStart, n);// QualityAssesser.similarity(comp, compType, c, n);
                //System.out.println("Similarity between cluster " + c.id + " to node " + n.getIdentifier() + " is " + similarity);
                if (similarity >= least) {
                    //System.out.println("added");
                    c.add(n.node, n.net);
                    notInClusterIter.remove();
                    nodesInClusterTemp.add(n);
                }
            }
            for (NetNode n : nodesInCluster) {
                double similarity = neighComp.similarity(nStart, n);//QualityAssesser.similarity(comp, compType, c, n);
                //System.out.println("Similarity between cluster " + c.id + " to node " + n.getIdentifier() + " is " + similarity);
                if (similarity >= least) {
                    //System.out.println("added");
                    c.add(n);
                }
            }
            nodesInCluster.addAll(nodesInClusterTemp);
            nodesInClusterTemp.clear();
            clusters.add(c);
        }
        //System.out.println("\nGarbage collect...");
        System.gc();
        System.out.println("Number of clusters " + clusters.size());
        int sum = 0;
        //char i = 'a';
        int i = 0;
        Cluster single = new Cluster(new Vector<NetNode>());
        single.id = '?';
        List<Cluster> clusterList = new Vector<Cluster>(clusters);
        Collections.sort(clusterList, new Comparator<Cluster>() {
            public int compare(Cluster o1, Cluster o2) {
                return o2.size() - o1.size();
            }
        });
        char maxChar = readableCharacters[0];
        for (Iterator<Cluster> clusterIter = clusterList.iterator(); clusterIter.hasNext();) {
            Cluster c = clusterIter.next();
            sum += c.size();
            if (c.size() == 1 && joinSingleClusters) {
                clusterIter.remove();
                single.add(c.getRepNetNodes().iterator().next());
            } else {
                if (i < readableCharacters.length) {
                    c.id = readableCharacters[i++];
                    if (c.id > maxChar) maxChar = c.id;
                } else {
                    c.id = ++maxChar;
                }
            }
            //System.out.println(c);
        }
        if (joinSingleClusters && single.size() > 0) {

            int index = Collections.binarySearch(clusterList, single, new Comparator<Cluster>() {
                public int compare(Cluster o1, Cluster o2) {
                    return o2.size() - o1.size();
                }
            });

            if (index < 0) {
                clusterList.add((-index) - 1, single);
            } else {
                clusterList.add(index, single);
            }

        }
        System.out.println("Total number of nodes: " + sum + ", factor: " + (double) sum / (double) nodesInCluster.size());

        return clusterList;
    }


}
