package nemi.compare;

import cytoscape.CyNetwork;
import giny.model.Node;
import nemi.NetNode;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

public class MetricCompare extends ProteinCompare {
    public static String name = "Metric comparer";
    private ProteinCompare comp;
    private HashMap<String, HashMap<String, Double>> similarityScore;
    private CyNetwork network;


    public MetricCompare(ProteinCompare comp, CyNetwork network) {
        this.comp = comp;
        this.network = network;
        similarityScore = new HashMap<String, HashMap<String, Double>>();
    }


    public String getName() {
        return this.name;
    }

    /**
     * Makes any similarity metric by adding a new dimension to the non-metric
     * space.
     * Note: if the similarity score already is metric there will be no change
     * to the similarity values
     */
    public void makeMetric() {
        int nodeCount = network.getNodeCount();
        // Create nodelist
        Vector<Node> nodeList = new Vector<Node>(nodeCount);
        for (Iterator<Node> nodeIter = network.nodesIterator(); nodeIter
                .hasNext();) {
            nodeList.add(nodeIter.next());
        }
        // N**3 iterator for looking at every possible triangle
        double minD = Double.MIN_VALUE;
        double ij, sij, ik, sik, jk, sjk;
        Node iNode, jNode, kNode;

        //sqrt(ij*ij) > sqrt(ik*ik) + sqrt(jk*jk)
        // implies that there exists a D such that
        // sqrt(ij*ij + D) = sqrt(ik*ik + D) + sqrt(jk*jk + D)

        //First we find this D by looking at ALL triangles  (N**3)
        for (int i = 0; i < nodeCount; i++) {
            iNode = nodeList.get(i);
            for (int j = i + 1; j < nodeCount; j++) {
                jNode = nodeList.get(j);
                ij = comp.similarity(new NetNode(network, iNode), new NetNode(network, jNode));
                sij = ij * ij;
                for (int k = j + 1; k < nodeCount; k++) {
                    kNode = nodeList.get(k);
                    ik = comp.similarity(new NetNode(network, iNode), new NetNode(network, kNode));
                    sik = ik * ik;
                    jk = comp.similarity(new NetNode(network, jNode), new NetNode(network, kNode));
                    sjk = jk * jk;
                    // check if triange inequality isn't satisfied
                    if (ij > ik + jk) {

                        minD = Math
                                .max(constantShiftSolve(sij, sik, sjk), minD);
                    }
                    if (ik > ij + jk) {
                        minD = Math
                                .max(constantShiftSolve(sik, sij, sjk), minD);
                    }
                    if (jk > ik + ij) {
                        minD = Math
                                .max(constantShiftSolve(sjk, sik, sij), minD);

                    }
                }
            }
        }
        double maxNewSim = Double.MIN_VALUE;
        String iId, jId;
        double newIJ;
        //then we calculate the new metric similarity value
        for (int i = 0; i < nodeCount; i++) {
            iNode = nodeList.get(i);
            iId = iNode.getIdentifier();
            for (int j = 0; j < nodeCount; j++) {
                jNode = nodeList.get(j);
                jId = jNode.getIdentifier();
                ij = comp.similarity(new NetNode(network, iNode), new NetNode(network, jNode));
                newIJ = Math.sqrt(ij * ij + minD);
                maxNewSim = Math.max(newIJ, maxNewSim);
                if (iId.compareTo(jId) < 0) {
                    if (!similarityScore.containsKey(iId)) {
                        similarityScore.put(iId, new HashMap<String, Double>());
                    }
                    Map<String, Double> m = similarityScore.get(iId);
                    m.put(jId, newIJ);
                } else {
                    if (!similarityScore.containsKey(jId)) {
                        similarityScore.put(jId, new HashMap<String, Double>());
                    }
                    Map<String, Double> m = similarityScore.get(jId);
                    m.put(iId, newIJ);
                }

            }

        }
        //Normalize the similarity between 0 and 1
        for (Map.Entry<String, HashMap<String, Double>> entry : similarityScore
                .entrySet()) {
            HashMap<String, Double> map1 = entry.getValue();
            for (Map.Entry<String, Double> entry2 : map1.entrySet()) {
                entry2.setValue(entry2.getValue() / maxNewSim);
            }

        }
    }

    /**
     * Finds a positive D such that sqrt(x + D) = sqrt(y + D) + sqrt(z + D)
     *
     * @param x squared distance
     * @param y squared distance
     * @param z squared distance
     * @return positive D
     */
    private double constantShiftSolve(double x, double y, double z) {
        double a = -3;
        double b = -2 * (x + y + z);
        double c = (x - y - z) * (x - y - z) - 4 * y * z;
        double temp = b * b - 4 * a * c;
        double d1, d2;
        if (temp >= 0) {
            d1 = (-b + Math.sqrt(temp)) / (2 * a);
            d2 = (-b - Math.sqrt(temp)) / (2 * a);
            return Math.max(d1, d2);

        }
        return -1;

    }

    public double similarity(NetNode nn1, NetNode nn2) {
        String n1 = nn1.node.getIdentifier();
        String n2 = nn2.node.getIdentifier();
        if (n1.compareTo(n2) < 0) {
            Map<String, Double> m = similarityScore.get(n1);
            return m.get(n2);
        } else {
            Map<String, Double> m = similarityScore.get(n2);
            return m.get(n1);
        }

    }

}
