package nemi.compare;

import cytoscape.CyNetwork;
import giny.model.Node;
import nemi.NetNode;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: kristost
 * Date: 26.mai.2006
 * Time: 09:49:21
 * <p/>
 * A comparer based on transcription factor (TF) activation levels. For the two nodes to be compared,
 * we find all common TFs and the corresponding activation level, for example: <br/>
 * A - tf1 - 0.1 <br/>
 * A - tf2 - 0.2 <br/>
 * B - tf1 - 0.3 <br/>
 * B - tf2 - 0.4 <br/>
 * If we want to calculate similarity between protein A & B, the score is: (0.1*0.3)+(0.2*0.4)
 */
public class TransFacCompare extends ProteinCompare {
    public final static String name = "Transcription factor comparer";
    public final static boolean requiresGO = false;
    Map<String, Double> scores = new HashMap<String, Double>();

    public String getName() {
        return this.name;
    }

    public String toString() {
        return name;
    }

    /**
     * Create a new comparer for all nodes in the provided nets.
     *
     * @param nets
     */
    public TransFacCompare(Collection<CyNetwork> nets) {
        Map<String, Map<String, Double>> p2tf2pval = nemi.Toolkit.readBindings();
        Map<String, Set<String>> id2tf = new HashMap<String, Set<String>>();
        Set<String> allProts = new HashSet<String>();
        for (CyNetwork net : nets) {
            for (Iterator<Node> nodeIter = net.nodesIterator(); nodeIter.hasNext();) {
                Node n = nodeIter.next();
                String id = n.getIdentifier();
                allProts.add(id);
                if (!p2tf2pval.containsKey(id)) id2tf.put(id, new HashSet<String>());
                if (!id2tf.containsKey(id))
                    id2tf.put(id, new HashSet<String>(p2tf2pval.get(id).keySet()));
            }

        }
        List<String> protList = new Vector<String>(allProts);
        Collections.sort(protList);
        // calc scores for all node-pairs
        double maxScore = -Double.MAX_VALUE;
        double minScore = Double.MAX_VALUE;
         int zeros = 0;
        for (int i = 0; i < protList.size(); i ++) {
            String p1 = protList.get(i);
            if (!p2tf2pval.containsKey(p1)) continue;
            for (int j = i + 1; j < protList.size(); j++) {
                String p2 = protList.get(j);
                if (!p2tf2pval.containsKey(p2)) continue;


                double score = 0.0;
                for (String s1 : id2tf.get(p1))
                    for (String s2 : id2tf.get(p2)) {
                        if (!s1.equals(s2)) continue;
                        if (p2tf2pval.get(p1).get(s1) == 0.0 || p2tf2pval.get(p2).get(s2) == 0.0) {
                            zeros ++;
                            continue;
                        }

                        //score += Math.log(p2tf2pval.get(p1).get(s1)) * Math.log(p2tf2pval.get(p2).get(s2));
                        score += (1-p2tf2pval.get(p1).get(s1)) * (1-p2tf2pval.get(p2).get(s2));

                    }
                scores.put(p1 + ":" + p2, score);
                maxScore = Math.max(score, maxScore);
                minScore = Math.min(score, minScore);
            }
        }
        for (Map.Entry<String, Double> entry : scores.entrySet()) {
            double d = (entry.getValue() -minScore) / (maxScore-minScore);
            entry.setValue(d);
        }
        System.out.println("" + zeros + " protein-protein combinations couldn't be evaluated through transfaccomparer.");
        //double minlog = Math.log10(minScore/maxScore);
        /*
        // normalize scores
        double scaling = 2;
        double newMax = -Double.MAX_VALUE;
        double newMin = Double.MAX_VALUE;
        for (Map.Entry<String, Double> entry : scores.entrySet()) {
            double d = Math.log(entry.getValue() + scaling) / Math.log(maxScore + scaling);
            newMax = Math.max(d, newMax);
            newMin = Math.min(d, newMin);
            entry.setValue(d);
        }

        for (Map.Entry<String, Double> entry : scores.entrySet()) {
            double d =( entry.getValue() -newMin) / (newMax-newMin);
            entry.setValue(d);
        } */
    }

    /**
     * @param n1
     * @param n2
     * @return The similarity for the two provided nodes.
     */
    public double similarity
            (NetNode
                    n1, NetNode
                    n2) {
        String p1 = n1.node.getIdentifier();
        String p2 = n2.node.getIdentifier();
        if (p1.equals(p2)) return 1.0;
        String key = (p1.compareTo(p2) < 0) ? p1 + ":" + p2 : p2 + ":" + p1;

        if (scores.containsKey(key)) return scores.get(key);

        return 0;
    }

}
