package nemi.gspan;

import cytoscape.CyNetwork;
import giny.model.Node;

import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: kristost
 * Date: 30.mar.2006
 * Time: 12:09:33
 *
 * A gSpan dfs node in the dfs search tree. Contains a motif and a set of instances.
 */
final class DfsNode implements Iterable<SubgraphInstance>, Comparable<DfsNode> {
    public final DfsMotif motif;
    public final HashSet<SubgraphInstance> instances = new HashSet<SubgraphInstance>();

    public DfsNode(DfsMotif motif) {
        this.motif = motif;
    }

    public void addInstance(SubgraphInstance i) {
        if (i != null)
            instances.add(i);
    }

    public void addAllInstances(Collection<SubgraphInstance> list) {
        for (SubgraphInstance instance : list) {
            if (instance != null)
                instances.add(instance);
        }
    }

    public int size() {
        return instances.size();
    }

    public int getNofNetworks(){
        Set<CyNetwork> count = new HashSet<CyNetwork>();
        for (SubgraphInstance instance : instances) {
            count.add(instance.net);

        }
        return count.size();
    }


    /**
     * For this node's motif and the corresponding instances, find all frequent forward children dfsNodes.
     * @param expandFrom Which node to expand from, using discovery number.
     * @param minsup Minimal support treshold.
     * @param multi Single or multiple networks.
     * @return All frequent dfsNode children of this node. Put in a sorted map, sorted on the new nodes' labels.
     */
    public SortedMap<Character, DfsNode> getFreqForwardChildren(int expandFrom, int minsup, boolean multi) {

        // counting the number of graphs each label can generate
        Map<Character, Integer> lblCount = new HashMap<Character, Integer>();
        Map<Character, HashSet<CyNetwork>> netset = new HashMap<Character, HashSet<CyNetwork>>();
        // save candidate nodes for each instance, mapped on lbls
        Map<Character, Map<SubgraphInstance, Set<Node>>> lbl2Graph2Nodes
                = new HashMap<Character, Map<SubgraphInstance, Set<Node>>>();

        for (SubgraphInstance inst : instances) {
            Collection<Node> neighs = inst.getUndiscoveredNeighs(expandFrom);
            for (Node neigh : neighs) {

                // increase support for this neigh nodes' label
                char lbl = neigh.getIdentifier().charAt(0);
                //Check condition 1 in smin test
                if (isBetterThanFirst(lbl, this.motif.labels.get(expandFrom), motif)) {
                    continue;
                }

                if (!lblCount.containsKey(lbl)){
                    lblCount.put(lbl, 0);
                    netset.put(lbl, new HashSet<CyNetwork>());
                }

                lblCount.put(lbl, lblCount.get(lbl) + 1);

                netset.get(lbl).add(inst.net);

                // new map
                if (!lbl2Graph2Nodes.containsKey(lbl)) {
                    lbl2Graph2Nodes.put(lbl, new HashMap<SubgraphInstance, Set<Node>>());
                }

                // new set for this instance
                if (!lbl2Graph2Nodes.get(lbl).containsKey(inst)) {
                    lbl2Graph2Nodes.get(lbl).put(inst, new HashSet<Node>());
                }

                // map label to instance to neigh set
                lbl2Graph2Nodes.get(lbl).get(inst).add(neigh);
            }
        }
        // for frequent labels, create new instances with frequent labels
        SortedMap<Character, DfsNode> ret = new TreeMap<Character, DfsNode>();
        for (Map.Entry<Character, Integer> counts : lblCount.entrySet()) {

            char lbl = counts.getKey();
            // creating the new code for the child instances
            DfsMotif childMotif = motif.addForward(expandFrom, lbl);

            // creating empty collection of instances
            DfsNode childDfsNode = new DfsNode(childMotif);
            ret.put(lbl, childDfsNode);

            if ((counts.getValue() >= minsup && !multi ) || (netset.get(counts.getKey()).size() >= minsup && multi)) {
                Map<SubgraphInstance, Set<Node>> parentMap = lbl2Graph2Nodes.get(lbl);
                for (SubgraphInstance inst : parentMap.keySet()) {
                    Set<Node> childNodes = parentMap.get(inst);
                    for (Node childNode : childNodes) {
                        /* checked in the instance, when getting neighs if(inst.nodes.contains(childNode))
                            continue;
                            */
                        childDfsNode.addInstance(
                                inst.createChildForward(childNode, expandFrom));
                    }
                }
            }
        }

        return ret;
    }

    /**
     * @param label1
     * @param label2
     * @param motif
     * @return Whether an edge containing the two provided labels is better than this motif's first edge.
     * If it is, this dfsNode may be safely discarded from the search tree.
     */
    private static boolean isBetterThanFirst(char label1, char label2, DfsMotif motif) {
        DfsCodeEdge firstEdge = motif.code.get(0);
        char temp;
        if (label1 > label2) {
            temp = label1;
            label1 = label2;
            label2 = temp;
        }
        if (firstEdge.l1 >label1) {
            return true;
        } else if(firstEdge.l1==label1){
            return firstEdge.l2>label2;
        }
        return false;
    }
    /*

    // Gets frequent children of this dfsnode, extending backward edges. Not used in the gSpan plugin,
    // but the implementation is tested successfully. To use this, subgraph instances must have a list of
    // nodes, not only the last node and a parent reference; or parents must be traversed in order to get the
    // correct backward nodes in the instance.

    public SortedMap<Integer, DfsNode> getFrequentBackward(int minSup, boolean multi) {
        int lastNode = instances.iterator().next().nodes.size() - 1;
        SortedMap<Integer, DfsNode> backInstances = new TreeMap<Integer, DfsNode>();
        char lastChar = motif.labels.get(lastNode);
        //int goto = 1;
        // System.out.println("DfsNode getFrequentBackward(): ");


        Vector<Integer> rightPath =  motif.getRightmostPathIxs();
        //remove two last because they are not interesting
        rightPath.remove(rightPath.size()-1);
        rightPath.remove(rightPath.size()-1);
        iterateRightmost:
        for (Integer j : rightPath) {
            if (j == lastNode) break;
            for (DfsCodeEdge e : motif) {
                //does this edge already exist?
                if(e.i == lastNode && e.j == j) continue iterateRightmost;
                if(e.j == lastNode && e.i == j) continue iterateRightmost;
                // Check for condition 2 in smin test:
                // no edge connected to i can be larger than new j,i
                if (e.i != j || !e.forward) continue;
                //first check if this edge exists
                char ichar = motif.labels.get(e.j);
                if (lastChar < ichar) {
                    // System.out.println("found smaller, pruning (pule): " + lastChar + "<" + ichar);
                    continue iterateRightmost;

                }
            }

            char jchar = motif.labels.get(j);
            if (isBetterThanFirst(lastChar, jchar, motif)) continue;

            // adding new instances for return
            Vector<SubgraphInstance> childInsts = new Vector<SubgraphInstance>();
            HashSet<CyNetwork> nets = new HashSet<CyNetwork>();
            for (SubgraphInstance inst : instances) {
                Node n1 = inst.nodes.get(lastNode);
                Node n2 = inst.nodes.get(j);

                boolean edgeExists = Toolkit.getNodeNeighborMap(inst.network).get(n1).contains(n2);
                if (edgeExists ) {//&& !inst.containsEdge(network, n1, n2)) {
                    childInsts.add(inst);
                    nets.add(inst.network);
                }
            }

            // cleaning for support
            if ((childInsts.size() >= minSup && !multi) || nets.size()>=minSup && multi) {
                DfsMotif motifWback = this.motif.addBackward(j);
                if (motifWback != null) {
                    DfsNode newDfsNode = new DfsNode(motifWback, this);
                    for(SubgraphInstance inst: childInsts){
                        Node n1 = inst.nodes.get(lastNode);
                        Node n2 = inst.nodes.get(j);
                        //Edge newEdge = Toolkit.getEdge(inst.network, n1, n2);
                        SubgraphInstance newInst = new SubgraphInstance(new Vector<Node>(inst.nodes), inst.network);// inst.clone();
                        newDfsNode.addInstance(newInst);
                    }
                    //newDfsNode.addAllInstances(childInsts);
                    backInstances.put(j, newDfsNode);
                }
            }
        }

        return backInstances;
    }



    public List<Node> getRightmostPathNodes(SubgraphInstance inst, DfsMotif motif) {
        // assuming the nodeInsts are rightmost only
        // return nodeInsts;

        Vector<Integer> ixs = motif.getRightmostPathIxs();

        List<Node> rm = new Vector<Node>();
        for (Integer ix : ixs) {
            rm.add(inst.nodes.get(ix));
        }
        return rm;
    }
      */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        //sb.append("Printing a DfsNode, containing code.. \n");
        sb.append(this.motif.toString()).append(". ");
        sb.append("Match in ");
        sb.append(this.getNofNetworks()).append(" networks, ");
        sb.append(this.instances.size()).append(" instances,");
        sb.append(motif.labels.size()).append(" nodes");
       // sb.append( instances.size() + " instances");


        //sb.append(motif.code.size() + "edges"); //we have removed backedges
        /*if (instances.size() < 5)
            for (SubgraphInstance instance : instances) {
                sb.append(instance.toString());
                sb.append("\n");
            }*/
        return sb.toString();
    }

    public Iterator<SubgraphInstance> iterator() {
        return instances.iterator();
    }

    /**
     * Comparing dfsNodes is comparing their corresponding motifs.
     */
    public int compareTo(DfsNode o) {
        return this.motif.compareTo(o.motif);
    }

}
