package nemi.gspan;

import cytoscape.CyEdge;
import cytoscape.CyNetwork;
import cytoscape.Cytoscape;
import cytoscape.giny.FingCyNetwork;
import cytoscape.data.Semantics;
import giny.model.Edge;
import giny.model.Node;
import giny.model.GraphPerspective;
import nemi.Toolkit;
import nemi.NetNode;
import nemi.cluster.Cluster;
import nemi.cluster.Clusterer;
import nemi.gui.NeMiFinalView;

import javax.swing.*;
import java.util.*;

/**
 * User: kristost
 * Date: 15.mar.2006
 * Time: 11:34:40
 * <p/>
 * A Cytoscape implementation of the gSpan graph mining algorithm. Mines for pattern in single
 * or multiple graphs.
 * <p/>
 * gSpan: graph-based substructure pattern mining
 * X Yan, J Han - Data Mining, 2002. ICDM 2002. Proceedings. 2002 IEEE ?, 2002
 */
public final class GSpanMining {

    private int graphSize;
    private final double bound = 0.5;
    private final NeMiFinalView.StatusInfo status;
    private boolean running = true;
    private int overlapCount = 0;

    public GSpanMining(NeMiFinalView.StatusInfo status, int overlapcount) {
        this.status = status;
        running = true;
        this.overlapCount = overlapcount;
        //noinspection unchecked
        List<CyNetwork> netset = new Vector<CyNetwork>((Set<CyNetwork>) Cytoscape.getNetworkSet());
        for (CyNetwork cyNetwork : netset) {
            if (cyNetwork.getTitle().startsWith(Toolkit.NetPrefix)) {
                try {
                    Cytoscape.destroyNetwork(cyNetwork);
                } catch (Exception e) {
                    System.err.println("Unable to destroy older nemi nets");
                    // e.printStackTrace();
                }
            }
        }
    }

    /**
     * Start mining for connected subgraphs in a collection of networks. Support is counted as the
     * number of occurences any motif has to be found in.
     *
     * @param cl                      All nodes need to be clustered, used for node labeling.
     * @param startNets               Collection of initial networks.
     * @param minsup                  The number of networks any motif must occur in.
     * @param explNet2origNet         Used for memoization and potential speedup. You may provide an empty map.
     * @param string2origNet2origNode Used for memoization and potential speedup. You may provide an empty map.
     * @return A list of DfsResults, used by the DfsResults class.
     */
    public List<DfsResult> subGraphMiningMulti(Collection<Cluster> cl, Collection<CyNetwork> startNets,
                                               int minsup, Map<CyNetwork, CyNetwork> explNet2origNet,
                                               Map<String, Map<CyNetwork, Node>> string2origNet2origNode,
                                               int maxsize) {
        List<Cluster> preparedCls = prepareClusters(cl, minsup);
        // List<CyNetwork> explodedNets = new Vector<CyNetwork>();  aldri brukt

        HashMap<DfsMotif, Collection<SubgraphInstance>> nodes = new HashMap<DfsMotif, Collection<SubgraphInstance>>();

        Map<DfsMotif, Collection<CyNetwork>> motif2Networks = new HashMap<DfsMotif, Collection<CyNetwork>>();
        status.setIndeterminate(true);
        for (CyNetwork network : startNets) {
            status.setProgressText("Preparing clusters for " + nemi.Toolkit.networkToString(network));
            CyNetwork newNet = createNetwork(network, preparedCls, string2origNet2origNode);
            if (!running) return null;
            explNet2origNet.put(newNet, network);
            // explodedNets.add(newNet);
            status.setProgressText("Creating initcodes for " + nemi.Toolkit.networkToString(network));
            Set<DfsNode> dfsNodes = createInitCodes(newNet);
            if (!running) return null;
            for (DfsNode dfsNode : dfsNodes) {
                if (!running) return null;
                if (!nodes.containsKey(dfsNode.motif)) {
                    nodes.put(dfsNode.motif, new HashSet<SubgraphInstance>());
                    motif2Networks.put(dfsNode.motif, new HashSet<CyNetwork>());
                }
                for (SubgraphInstance instance : dfsNode.instances) {
                    try {
                        nodes.get(dfsNode.motif).add(instance);
                        motif2Networks.get(dfsNode.motif).add(network);
                    } catch (Exception e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        System.err.println("feilen er: " + instance.getClass() + ", " + instance);
                    }
                }
            }
            if (!running) return null;
        }
        status.setProgressText("Merging initcodes");
        List<DfsNode> dfsNodes = new Vector<DfsNode>();
        for (Map.Entry<DfsMotif, Collection<SubgraphInstance>> entry : nodes.entrySet()) {
            DfsNode newNode = new DfsNode(entry.getKey());
            newNode.addAllInstances(entry.getValue());
            dfsNodes.add(newNode);
        }

        Collections.sort(dfsNodes);
        Set<DfsResult> freq = new HashSet<DfsResult>();

        this.graphSize = 2;

        status.setMinimum(0);
        status.setMaximum(dfsNodes.size());
        status.setIndeterminate(false);
        int count = 0;
        for (DfsNode node : dfsNodes) {
            status.setValue(count++);
            status.setProgressText("DFS from " + node.motif.toString());
            if (motif2Networks.get(node.motif).size() >= minsup)
                subGraphMining(node, freq, minsup, true, maxsize);
        }
        status.setValue(0);
        status.setProgressText("");
        return new Vector<DfsResult>(freq);
    }

    /**
     * Start mining for connected subgraphs in a single network.
     */
    public List<DfsResult> subGraphMining(Collection<Cluster> cl, CyNetwork startNetwork, int minsup,
                                          Map<CyNetwork, CyNetwork> explNet2origNet,
                                          Map<String, Map<CyNetwork, Node>> string2origNet2origNode, int maxsize) {

        status.setIndeterminate(true);
        status.setProgressText("Preparing clusters.");
        List<Cluster> preparedCls = prepareClusters(cl, minsup);
        status.setProgressText("Creating exploded network.");
        if (!running) return null;
        CyNetwork explodedNet = this.createNetwork(startNetwork, preparedCls, string2origNet2origNode);
        if (!running) return null;
        explNet2origNet.put(explodedNet, startNetwork);
        Vector<CyNetwork> netlist = new Vector<CyNetwork>();
        netlist.add(explodedNet);
        status.setProgressText("Cleaning exploded network.");
        cleanNet(netlist, minsup);
        if (!running) return null;
        status.setProgressText("Creating inital codes.");
        SortedSet<DfsNode> instances = createInitCodes(explodedNet);
        Set<DfsResult> freq = new HashSet<DfsResult>();
        // creating tree out of all one-edge instances
        this.graphSize = 2;
        status.setIndeterminate(false);
        status.setMinimum(0);
        status.setMaximum(instances.size());
        int count = 1;
        for (DfsNode dfsNode : instances) {
            status.setValue(count++);
            status.setProgressText("DFS from " + dfsNode.motif.toString());
            status.repaint();
            //System.out.println("\n________________________________\nMining: " + dfsNode);
            if (!running) return null;
            subGraphMining(dfsNode, freq, minsup, false, maxsize);
        }
        status.setProgressText("");
        status.setValue(0);

        System.out.println("\n\n\nFinished mining\n\n\n");


        return new Vector<DfsResult>(freq);
    }

    private List<Cluster> prepareClusters(Collection<Cluster> cl, int minsup) {
        List<Cluster> clusters = new Vector<Cluster>();

        //remove small clusters
        List<Cluster> removedClusters = new Vector<Cluster>();
        for (Cluster c : cl)
            if (c.getRepNodes().size() >= minsup) {
                clusters.add(c);

            } else {
                removedClusters.add(c);
            }

        //sort clusters, descending size
        Collections.sort(clusters, new Comparator<Cluster>() {
            public int compare(Cluster o1, Cluster o2) {
                return o2.size() - o1.size();
            }
        });

        //relabeling clusters; sorted
        List<Character> charList = new Vector<Character>();
        for (char c : Clusterer.readableCharacters) {
            charList.add(c);
        }
        Collections.sort(charList);
        int i = 0;
        System.out.print("{");
        char maxLabel = charList.get(charList.size() - 1);
        for (Cluster cluster : clusters) {
            if (i >= charList.size()) {
                cluster.id = ++maxLabel;
            } else {
                cluster.id = charList.get(i++);
            }
            System.out.print("'" + cluster.id + "', ");
        }
        while (i < charList.size()) {
            System.out.print("'" + charList.get(i++) + "', ");
        }
        System.out.println("}");
        for (Cluster c : removedClusters) {
            if (i >= charList.size()) {
                c.id = ++maxLabel;
            } else {
                c.id = charList.get(i++);
            }
        }

        return clusters;
    }

    /**
     * Explode the current network. Uses the given clusters to label all nodes, creating a net network
     * where all nodes occuring in several clusters are duplicated.
     *
     * @param fromOld                 Original network
     * @param clusters                All nodes clustered.
     * @param string2origNet2origNode Map from cluster label, to the original net, to the original node.
     * @return A new cyNetwork; an exploded version of the original.
     */
    private CyNetwork createNetwork(CyNetwork fromOld, Collection<Cluster> clusters, Map<String, Map<CyNetwork, Node>> string2origNet2origNode) {

        CyNetwork network = Cytoscape.createNetwork(Toolkit.NetPrefix + fromOld.getTitle() + "(nemi)");
        Map<Node, Set<Cluster>> node2Clusters = this.createNodeClusterMap(clusters);
        HashMap<Node, Set<Node>> old2New = new HashMap<Node, Set<Node>>();

        // collect surviving nodes
        Set<Node> nodes = new HashSet<Node>();
        for (//noinspection unchecked
                Iterator<Node> nIter = (Iterator<Node>) fromOld.nodesIterator(); nIter.hasNext();) {
            nodes.add(nIter.next());
        }

        System.out.println("Network nodes: " + fromOld.getNodeCount() + ", clusterNodes: " + nodes.size());

        // add a new nodes for each cluster in the new network
        for (Node node : nodes) {
            Set<Cluster> coccs = node2Clusters.get(node);
            if (coccs == null)
                continue;
            Set<Node> newNodes = new HashSet<Node>();
            old2New.put(node, newNodes);
            Map<CyNetwork, Node> origNet2origNode = string2origNet2origNode.get(node.getIdentifier());
            if (origNet2origNode == null) {
                origNet2origNode = new HashMap<CyNetwork, Node>();
                string2origNet2origNode.put(node.getIdentifier(), origNet2origNode);
            }
            origNet2origNode.put(fromOld, node);
            for (Cluster cluster : coccs) {
                Node newNode = Cytoscape.getCyNode(cluster.id + "." + node.getIdentifier(), true);// new CyNode(network.getRootGraph(), ix++);
                // newNode.setIdentifier(cluster.id + "." + node.getIdentifier());
                network.addNode(newNode);
                newNodes.add(newNode);
                if(!running) return fromOld;

            }
        }

        // add all edges beween all nodes
        HashSet<String> edgeNameSet = new HashSet<String>();
        for (//noinspection unchecked
                Iterator<Edge> it = fromOld.edgesIterator(); it.hasNext();) {
            Edge e = it.next();
            Set<Node> fromNew = old2New.get(e.getTarget());
            Set<Node> toNew = old2New.get(e.getSource());
            if (fromNew == null || toNew == null) continue;
            for (Node fromNode : fromNew) {
                for (Node toNode : toNew) {
                    if(!running) return fromOld;
                    String sourceName = fromNode.getIdentifier();
                    String targetName = toNode.getIdentifier();
                    String label = "pp";
                    String edgeName = sourceName + " (" + label + ") " + targetName;
                    int duplicate_count = 1;
                    while (!edgeNameSet.add(edgeName)) {
                        edgeName = sourceName + " (" + label + ") " + targetName
                                + "_" + duplicate_count;
//					edgeName = sourceName + " " + label + " " + targetName
//						+ "_" + duplicate_count;

                        duplicate_count += 1;
                    }

                    CyEdge newEdge = Cytoscape.getRootGraph().getEdge(edgeName);

                    //CyEdge newEdge = Cytoscape.getCyEdge(fromNode, toNode,  Semantics.INTERACTION, e.getIdentifier(), true);

                    if (newEdge == null) {
                        Node node_1 = Cytoscape.getCyNode(sourceName);
                        Node node_2 = Cytoscape.getCyNode(targetName);
                        newEdge = Cytoscape.getCyEdge(node_1, node_2,
                                Semantics.INTERACTION, label, true);
                    }

                    // System.out.println("source: " + newEdge.getSource().getIdentifier() + ", target: " + newEdge.getTarget().getIdentifier());
                    network.addEdge(newEdge);
                }
            }
        }

        return network;
    }

    /**
     * Clean the given network for infrequent edges and un-connected nodes.
     *
     * @param networks The networks to be cleaned.
     * @param minsup   Minimum edge support.
     */
    private  void cleanNet(Collection<CyNetwork> networks, int minsup) {
        // count edges
        Map<String, Vector<Edge>> countE = new HashMap<String, Vector<Edge>>();
        Map<String, Vector<CyNetwork>> countN = new HashMap<String, Vector<CyNetwork>>();
        for (CyNetwork network : networks)    {
            for (//noinspection unchecked
                    Iterator<Edge> it = network.edgesIterator(); it.hasNext();) {
                Edge e = it.next();
                String s1 = e.getSource().getIdentifier().substring(0, 1);
                String s2 = e.getTarget().getIdentifier().substring(0, 1);
                String s = (s1.compareTo(s2) < 0) ? s1 + s2 : s2 + s1;

                //Toolkit.p
                if (!countE.containsKey(s)) {
                    countE.put(s, new Vector<Edge>());
                    countN.put(s, new Vector<CyNetwork>());
                }
                Vector<Edge> edges = countE.get(s);
                Vector<CyNetwork> nets = countN.get(s);
                edges.add(e);
                nets.add(network);
            if(!running) return;
            }

        }

        // prune edges

        for (Map.Entry<String, Vector<Edge>> e : countE.entrySet()) {
            if (e.getValue().size() < minsup) {
                Vector<Edge> vecEdge = e.getValue();
                Vector<CyNetwork> vecNet = countN.get(e.getKey());
                for (int i = 0; i < vecEdge.size(); i++) {
                    if(!running) return;
                    CyNetwork net = vecNet.get(i);
                    Edge edge = vecEdge.get(i);
                    int edgeIx = net.getIndex(edge);
                    // int before = net.getAdjacentEdgeIndicesArray(net.getIndex(edge.getTarget()), true, true, true).length;
                    try {
                        net.hideEdge(edgeIx);
                    } catch (Exception e1) {
                        //e1.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                    // int after = net.getAdjacentEdgeIndicesArray(net.getIndex(edge.getTarget()), true, true, true).length;
                    //System.out.println("neighcount " + before + " / " + after);

                    //vecNet.get(i).removeEdge(vecNet.get(i).getIndex(vecEdge.get(i)), true);

                }
            }
        }


        for (CyNetwork network : networks) {
            // remove disconnected nodes
            for (//noinspection unchecked
                    Iterator<Node> it = network.nodesIterator(); it.hasNext();) {
                Node n = it.next();
                if (network.getDegree(n) == 0) {
                    try {
                        if(!running) return;
                        network.removeNode(network.getIndex(n), true);
                    } catch (Exception e) {
                        //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
            }
        }
    }

    /**
     * Creates all distinct codes with a single edge, return sorted.
     * All edges references to the actual CyNetwork edges through instances prop.
     *
     * @return All possible single edge codes
     */
    private  SortedSet<DfsNode> createInitCodes(CyNetwork network) {
        // mapping cluster-id's to edges, to keep only one instance of each
        HashMap<char[], DfsCodeEdge> initSet = new HashMap<char[], DfsCodeEdge>();
        // maps graphedges to graph instances
        HashMap<DfsCodeEdge, List<List<Node>>> ce2liliNo = new HashMap<DfsCodeEdge, List<List<Node>>>();


        for (//noinspection unchecked
                Iterator<Edge> ei = network.edgesIterator(); ei.hasNext();) {
            Edge e = ei.next();
            Node n1 = e.getSource();
            Node n2 = e.getTarget();
            char l1 = n1.getIdentifier().charAt(0);
            char l2 = n2.getIdentifier().charAt(0);
            char[] li = (l1 < l2) ? new char[]{l1, l2} : new char[]{l2, l1};
            DfsCodeEdge newEdge = new DfsCodeEdge(0, 1, li[0], li[1]);
            if (!initSet.containsKey(li)) {
                initSet.put(li, newEdge);
            }
            if (!ce2liliNo.containsKey(newEdge)) {
                ce2liliNo.put(newEdge, new Vector<List<Node>>());
            }
             if (!running) return new TreeSet<DfsNode>();
            Vector<Node> thisOcc = new Vector<Node>();
            if (l1 < l2) {
                thisOcc.add(n1);
                thisOcc.add(n2);
            } else {
                thisOcc.add(n2);
                thisOcc.add(n1);
            }
            ce2liliNo.get(newEdge).add(thisOcc);

            if (l1 == l2) {
                Vector<Node> thisOcc2 = new Vector<Node>();
                thisOcc2.add(n1);
                thisOcc2.add(n2);
                ce2liliNo.get(newEdge).add(thisOcc2);
            }
        }

        // creating codes out of edges.
        int instanceCount = 0;
        TreeSet<DfsNode> ret = new TreeSet<DfsNode>();
        for (DfsCodeEdge e : initSet.values()) {
            DfsMotif motif = new DfsMotif(e);
            DfsNode dfsNode = new DfsNode(motif);
            for (List<Node> inst : ce2liliNo.get(e)) {
                if(!running) return ret;
                //Set<Edge> edgeSet = new HashSet<Edge>();
                //Edge ef = Toolkit.getEdge(network, inst.get(0), inst.get(1));
                //edgeSet.add(ef);
                SubgraphInstance oneNode = new SubgraphInstance(inst.get(0), null, network);
                SubgraphInstance secNode = new SubgraphInstance(inst.get(1), oneNode, network);
                dfsNode.addInstance(secNode);
                instanceCount ++;
            }
            ret.add(dfsNode);
        }

        System.out.println("GSpanMining.createInitCodes() created " + ret.size() + " initial codes, " + ret.size()
                + " and " + instanceCount + " instances");

        return ret;
    }

    private static Map<Node, Set<Cluster>> createNodeClusterMap(Collection<Cluster> clusters) {
        Map<Node, Set<Cluster>> node2Clusters = new HashMap<Node, Set<Cluster>>();

        // maps nodes to clusters
        for (Cluster c : clusters) {
            for (Node n : c.getRepNodes()) {
                if (node2Clusters.containsKey(n))
                    node2Clusters.get(n).add(c);
                else {
                    Set<Cluster> newSet = new HashSet<Cluster>();
                    newSet.add(c);
                    node2Clusters.put(n, newSet);
                }
            }
        }
        return node2Clusters;
    }

    /**
     * Recursive method for the subgraph mining according to the article.
     *
     * @param s         The set of candidate subgraphs.
     * @param frequents The set of frequent subgraphs.
     * @param minsup    Minimum support
     * @param multi     Set if mining multiple networks.
     * @param maxsize   Maximum subgraph size.
     */
    private void subGraphMining(DfsNode s, Set<DfsResult> frequents, int minsup, boolean multi, int maxsize) {
        //if (counter++ > 100) return;


        if (!running) return;
        int nNodes = s.motif.labels.size();

        // saving instance if larger than already discovered, or add if equal

        if (nNodes > graphSize) {
            for (Iterator<DfsResult> iterator = frequents.iterator(); iterator.hasNext();) {
                DfsResult dfsNode = iterator.next();
                if (dfsNode.motif.labels.size() < graphSize * bound) {
                    iterator.remove();
                }
            }
            frequents.add(new DfsResult(s));
            graphSize = (nNodes);
        } else if (nNodes >= graphSize * bound) {
            frequents.add(new DfsResult(s));
        }
        Vector<DfsNode> C = new Vector<DfsNode>();
        if (nNodes >= maxsize) return;
        // collecting backward nodes
        // Vector<DfsNode> C = new Vector<DfsNode>(s.getFrequentBackward(minsup, multi).values());

        // .. and forwards
        for (int i : s.motif.getRightmostPathIxs()) {
            // check memory usage..
            double free = ((double) Runtime.getRuntime().freeMemory()) / (double) Runtime.getRuntime().totalMemory();
            if (free < 0.2) {
                System.runFinalization();
                System.gc();

                free = ((double) Runtime.getRuntime().freeMemory()) / (double) Runtime.getRuntime().totalMemory();
                if (free < 0.2 && !moreMemoryRequired) {
                    moreMemoryRequired = true;
                    Thread run = new Thread() {
                        public void run() {
                            JOptionPane.showMessageDialog(null, "Ran out of memory. Some results will be omitted",
                                    "Stack space full",
                                    JOptionPane.ERROR_MESSAGE);
                        }
                    };
                    run.start();

                }
                if (free < 0.2) break;

            }
            if (!running) return;
            SortedMap<Character, DfsNode> forwards = s.getFreqForwardChildren(i, minsup, multi);
            C.addAll(forwards.values());
        }

        Collections.sort(C);
        for (DfsNode sr : C) {
            if (!running) return;
            if (sr.compareTo(s) > 0 && sr.motif.code.size() > s.motif.code.size()) {
                if (multi) {
                    pruneDuplicates(sr);
                    if(sr.getNofNetworks()<minsup) continue;
                }  else {
                    if(!hasEnoughUniqeMatches(sr, minsup))continue;
                    if(sr.size()<minsup) continue;
                }
                if (sr.motif.isSmallest())
                    subGraphMining(sr, frequents, minsup, multi, maxsize);
                //else System.out.print(++npruned + " ");

            }
        }
    }

    private void pruneDuplicates(DfsNode sr) {
        for (Iterator<SubgraphInstance> sgIter = sr.iterator(); sgIter.hasNext();) {
            Set<String> instanceList = new HashSet<String>();
           SubgraphInstance temp = sgIter.next();
            while (temp != null) {

                String n = temp.net.hashCode() + "" + temp.lastNode.getIdentifier().substring(2);

                instanceList.add(n);
                temp = temp.parentInstance;
            }
            if (instanceList.size() < sr.motif.labels.size()) {
                sgIter.remove();
            }
        }
    }

    private boolean moreMemoryRequired = false;
    private int npruned = 0;

    /**
     * Decided whether the given DfsNode's instances have enough unique instances.
     * If all instances' nodes have some overlaps, the method will return true.
     * Also removes unwanted duplicates
     *
     * @param s
     * @param minsup
     * @return True if the given DfsNode has enough unique matches.
     */
    private boolean hasEnoughUniqeMatches(DfsNode s, int minsup) {

        Set<String> discovered = new HashSet<String>();
// if (multi) return s.getNofNetworks() >= minsup;
        int uniqueCount = 0;
        for (Iterator<SubgraphInstance> itSub = s.iterator(); itSub.hasNext();) {
            SubgraphInstance instance = itSub. next();
            int thisUnique = 0;
            Set<String> instanceList = new HashSet<String>();
            SubgraphInstance temp = instance;
            while (temp != null) {
                String n = temp.net.hashCode() + "" + temp.lastNode.getIdentifier().substring(2);
                if (!discovered.contains(n)) {
                    thisUnique++;
                }
                instanceList.add(n);
                temp = temp.parentInstance;
            }

            if (instanceList.size() == s.motif.labels.size()) {
                if (thisUnique >= instanceList.size() - this.overlapCount) {
                    discovered.addAll(instanceList);
                    uniqueCount++;
                }
            } else {
                itSub.remove();
            }
        }
        if (uniqueCount >= minsup) return true;
        return false;  //To change body of created methods use File | Settings | File Templates.
    }


    public void setRunning(boolean b) {
        running = b;
    }

    public List<CyNetwork> generateRandomNets(CyNetwork fromnet, int nnets, int nodecount) {

        GraphPerspective gp = fromnet.getGraphPerspective();
        Map<Node, Set<Node>> neighmap = Toolkit.getNodeNeighborMap(fromnet);
        Random r = new Random(0);
        Node[] allNodes = new Node[gp.getNodeCount()];
        int i = 0;
        for (Iterator<Node> nit = gp.nodesIterator(); nit.hasNext();) {
            allNodes[i++] = nit.next();
        }

        List<CyNetwork> newnets = new Vector<CyNetwork>();
        for (int j = 0; j < nnets; j++) {
            try {
                CyNetwork newnet = Cytoscape.createNetwork("gSpanTestNet" + j);
                Node startnode = allNodes[(int) (r.nextDouble() * gp.getNodeCount())];
                newnet.addNode(startnode);
                for (i = 0; i < nodecount; i++) {
                    Node rnode = newnet.getNode(
                            newnet.getNodeIndicesArray()[
                                    (int) (r.nextDouble() * newnet.getNodeIndicesArray().length)]
                    );
                    Node[] neighs = new Node[neighmap.get(rnode).size()];
                    neighmap.get(rnode).toArray(neighs);
                    if (neighs.length == 0) continue;
                    Node rneig = neighs[(int) (r.nextDouble() * neighs.length)];
                    Edge e = Toolkit.getEdge(fromnet, rnode, rneig);
                    newnet.addNode(rneig);
                    newnet.addEdge(e);
                }
                newnets.add(newnet);
            } catch (Exception e) {
            }
        }
        return newnets;
    }

    public List<CyNetwork> insertSubgraph
            (CyNetwork
                    supernet, List<CyNetwork> nets, CyNetwork
                    motif, int nin) {
        Random r = new Random(1);
        if (nin > nets.size()) return null;
        Set<Node> allmnodes = new HashSet<Node>();
        for (Iterator<Node> nit = motif.getGraphPerspective().nodesIterator(); nit.hasNext();)
            allmnodes.add(nit.next());
        Set<Edge> allmedges = new HashSet<Edge>();
        for (Iterator<Edge> nit = motif.getGraphPerspective().edgesIterator(); nit.hasNext();)
            allmedges.add(nit.next());
        Set<Edge> superedges = new HashSet<Edge>();
        for (Iterator<Edge> nit = supernet.getGraphPerspective().edgesIterator(); nit.hasNext();)
            superedges.add(nit.next());


        for (int i = 0; i < nin; i++) {
            CyNetwork grownet = nets.get(i);
            Set<Node> grownetnodes = new HashSet<Node>();
            Set<Edge> grownetedges = new HashSet<Edge>();
            for (Iterator<Node> nit = grownet.getGraphPerspective().nodesIterator(); nit.hasNext();)
                grownetnodes.add(nit.next());
            for (Iterator<Edge> eit = grownet.getGraphPerspective().edgesIterator(); eit.hasNext();)
                grownetedges.add(eit.next());

            // have commons, add remainding
            Set<Node> commons = new HashSet<Node>(allmnodes);
            if (commons.removeAll(grownetnodes)) {
                for (Node node : commons) {
                    for (Node neigh : Toolkit.getNeighbors(new NetNode(motif, node))) {
                        if (!grownetnodes.contains(neigh)) {
                            grownetnodes.add(neigh);
                            grownet.addEdge(Toolkit.getEdge(motif, node, neigh));
                            grownet.addNode(neigh);
                        }
                    }
                }
                for (Edge edge : allmedges) {
                    if (!grownetedges.contains(edge)) {
                        grownet.addEdge(edge);
                        grownetedges.add(edge);
                    }
                }
            }

            // didnt have commons
            else {
                // find edge connecting grownet to node, addall mnodes & edges
                Node grownode = grownet.getNode(
                        grownet.getNodeIndicesArray()[
                                (int) (r.nextDouble() * grownet.getNodeIndicesArray().length)]);
                Node motifnode = motif.getNode(motif.getNodeIndicesArray()[
                        (int) (r.nextDouble() * motif.getNodeIndicesArray().length)]);

                Edge newEdge = Cytoscape.getCyEdge(grownode, motifnode,
                        Semantics.INTERACTION, "pp", true);

                superedges.add(newEdge);

                // add connecting edge
                grownet.addEdge(newEdge);
                grownetedges.add(newEdge);
                for (Node node : allmnodes) {
                    if (!grownetnodes.contains(node)) {
                        grownetnodes.add(node);
                        grownet.addNode(node);
                    }
                }
                for (Edge edge : allmedges) {
                    if (!grownetedges.contains(edge)) {
                        grownetedges.add(edge);
                        grownet.addEdge(edge);
                    }
                }

            }
            // add all remaining edges
            for (Edge sedge : superedges) {
                Edge growedge = null;

                if (grownetnodes.contains(sedge.getSource()) &&
                        grownetnodes.contains(sedge.getTarget())) {
                    growedge = sedge;
                }
                if (grownetnodes.contains(sedge.getTarget()) &&
                        grownetnodes.contains(sedge.getSource())) {
                    growedge = sedge;
                }
                if (growedge != null) {
                    grownet.addEdge(growedge);
                    grownetedges.add(growedge);
                }
            }

        }

        return nets;
    }


}
