package nemi.motif;

import nemi.gui.MessagePane;
import nemi.gui.ViewableResults;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

public class AssignmentResults extends ViewableResults {

    private List<Assignment> resultMap;
    private Map<String, Assignment> uniqueAssignments;
    private int maxResults;
    private MotifFinder motifFinder;
    private MessagePane messagePane;

    public AssignmentResults(int maxResults, MessagePane messagePane) {
        super();
        // resultMap = Collections.synchronizedMap(new TreeMap<Assignment, Assignment>(new AssignmentComparator()));
        resultMap = Collections.synchronizedList(new Vector<Assignment>());
        this.messagePane = messagePane;
        this.maxResults = maxResults;
        uniqueAssignments = Collections.synchronizedMap(new HashMap<String, Assignment>());
    }

    public void setMotifFinder(MotifFinder mf) {
        this.motifFinder = mf;
    }

    /**
     * Flagging the selected assignments' nodes.
     *
     * @param e
     */
    public void valueChanged(ListSelectionEvent e) {
        //Ignore extra messages.
        if (e.getValueIsAdjusting()) return;


        ListSelectionModel lsm =
                (ListSelectionModel) e.getSource();
        if (!lsm.isSelectionEmpty()) {

            int selectedRow = lsm.getMinSelectionIndex();
            if (!lsm.isSelectedIndex(selectedRow)) selectedRow = lsm.getMaxSelectionIndex();
            if (selectedRow >= 0 && selectedRow < resultMap.size()) {
                Assignment ass = resultMap.get(selectedRow);
                motifFinder.setFlagMotifNodes(ass);
                motifFinder.setFlagNetworkNodes(ass);
                messagePane.print(ass.toString() + "\n");
            }
        }
    }

    public void stopRunning() {
        motifFinder.stopRunning();
    }

    /**
     * Overriding the tooltiptext
     */

    public String getToolTipText(int rowIndex) {
        return resultMap.get(resultMap.size() - rowIndex - 1).toHtml();
    }


    private double interestingLimit = -1;
    public void addAssignment(Assignment ass) {
        if (ass == null) return;
        if (resultMap.size() > maxResults * 2) {
            sortList();
            interestingLimit = Math.max(interestingLimit, resultMap.get(maxResults).score);
            resultMap = new Vector<Assignment>(resultMap.subList(0, maxResults));
        }

        if (uniqueAssignments.containsKey(ass.getNodeHashString())) {
            Assignment prevBest = uniqueAssignments.get(ass.getNodeHashString());
            if (ass.score > prevBest.score) {
                resultMap.remove(prevBest);
                resultMap.add(ass);
                uniqueAssignments.put(ass.getNodeHashString(), ass);
            }
        } else if (ass.score > interestingLimit) {
            resultMap.add(ass);
            uniqueAssignments.put(ass.getNodeHashString(), ass);
        }
        fireTableDataChanged();
    }

    public void saveToFile() {
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle("Choose file");
        fc.setCurrentDirectory(new File("."));
        JFrame saveFrame = new JFrame();
        saveFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        int result = fc.showSaveDialog(saveFrame);
        saveFrame.setVisible(false);
        if (result == JFileChooser.APPROVE_OPTION) {
            File saveto = fc.getSelectedFile();
            try {
                FileWriter out = new FileWriter(saveto);
                if (resultMap.size() == 0) out.write("No matches found.");
                for (Assignment assignment : resultMap) {
                    out.write(assignment.toString() + "\n");
                }
                out.close();
                System.out.println("Saved resultMap to " + saveto);
            } catch (IOException e) {
                System.out.println("Couldn't write to " + saveto);
            }
        }
    }

    public String toString() {
        if (resultMap.size() == 0) return "No matches found.";
        String s = "";
        for (Assignment assignment : resultMap) {
            s += assignment.toString() + "\n";
        }
        return s;
    }

    private void sortList() {
        Collections.sort(resultMap, new Comparator<Assignment>() {
            public int compare(Assignment o1, Assignment o2) {
                if (o1.score > o2.score) return -1;
                else if (o2.score > o1.score) return 1;
                return 0;
            }
        });
    }

    public synchronized void purgeLessScore(double score) {
        sortList();
        for (Iterator<Assignment> ait = resultMap.iterator(); ait.hasNext();){
            if (ait.next().score < score) {
                ait.remove();
            }
        }
        fireTableDataChanged();
    }


    public int getRowCount() {
        // threading is cool..
        if (resultMap == null) return 0;

        return Math.min(maxResults, resultMap.size());
    }

    public int getColumnCount() {

        return 2;
    }

    public String getColumnName(int columnIndex) {
        switch (columnIndex) {
            case 0:
                return "Assignment size";
            case 1:
                return "Assignment score";
        }
        return "THojo";
    }

    public Class<?> getColumnClass(int columnIndex) {
        switch (columnIndex) {
            case 0:
                return Integer.class;
            case 1:
                return Double.class;
            default:
                return String.class;
        }
    }

    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return false;
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        if (resultMap != null && resultMap.size() > rowIndex) {
            switch (columnIndex) {
                case 0:
                    return resultMap.get(rowIndex).getMotifNodes().size();
                case 1:
                    return resultMap.get(rowIndex).score;
            }
        }
        return null;
    }


    class AssignmentComparator implements Comparator {
        public int compare(Object o1, Object o2) {
            if (o1 instanceof Assignment && o2 instanceof Assignment) {
                double diff = (((Assignment) o1).score - ((Assignment) o2).score);
                if (diff == 0.0) return 0;
                else return diff < 0 ? -1 : 1;
            } else
                return 0;
        }
    }

    public static void main
            (String[] args) {
        // AssignmentResults ar = new AssignmentResults(100);
        // ar.saveToFile();
    }
}
