package nemi.motif;

import cytoscape.CyNetwork;
import cytoscape.Cytoscape;
import giny.model.Node;
import nemi.NetNode;
import nemi.Toolkit;
import nemi.gui.NeMiFinalView;
import nemi.compare.ProteinCompare;

import javax.swing.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA. User: kristost Date: 01.feb.2006 Time: 14:45:12
 * Performs a depth-first search, trying to match a given motif in a given network. All
 * motif nodes are searched against all network nodes. Only forward edges are considered,
 * that is, edges defining be best match between the motif and network's span trees.
 * <p/>
 * The algorithm uses the given comparer for scoring two node's similarity. The search bounds
 * on the given branch factor. With a branch factor of 1.0 the algorithm will perform a complete
 * search.
 */
public class MotifFinder extends Thread {
    private boolean DEBUG = true;
    private CyNetwork network;
    private CyNetwork motif;
    private nemi.compare.ProteinCompare comparer;
    private Map<Node, Set<Node>> motifNeighborMap;
    private Map<Node, Set<Node>> netNeighborMap;
    //private Set<Node> blackList;
    private boolean running = true;
    private nemi.motif.MotifFinderPanel motifFinderFrame;
    private AssignmentResults assResults;
    private double branchFactor; // bf = .8 means branch on everything smaller than ,8*max
    private double bestScore;
    private NeMiFinalView.StatusInfo status;
    // needed to disable flags
    boolean moreMemoryRequired = false;

    /**
     * Initialize the MotifFinder for finding the provided motif in the given network, using the
     * provided comparer.
     *
     * @param network    Network to find the given motif in
     * @param motif      Motif to find in the given network
     * @param comparer   Comparer, needed to compare node to node similarity
     * @param assResults Results are appended during the search
     * @param status
     */

    public MotifFinder(CyNetwork network, CyNetwork motif,
                       ProteinCompare comparer, double branchFactor,
                       nemi.motif.MotifFinderPanel motifFinderFrame, AssignmentResults assResults,
                       NeMiFinalView.StatusInfo status) {
        super();
        this.status = status;
        this.network = network;
        this.motif = motif;
        this.comparer = comparer;
        this.branchFactor = branchFactor;
        // System.out.println("similarity Treshold: " + branchFactor);

        this.motifNeighborMap = Toolkit.getNodeNeighborMap(motif);
        this.netNeighborMap = Toolkit.getNodeNeighborMap(network);
        this.motifFinderFrame = motifFinderFrame;
        this.assResults = assResults;

        assResults.setMotifFinder(this);
        this.start();


    }

    /**
     * Start the thread. Results are appended to assResults when found.
     */
    public void run() {
        final Assignment result = this.getMotifMatch();
        this.motifFinderFrame.searchFinished();
        if (result != null && result.score > 0) {
            Toolkit.permutationsMemo = new HashMap<String, List<List<Integer>>>();
            // System.out.println(result);
            // new ErrorFrame("Results", result.toString());
        }
        running = false;
    }

    public boolean isRunning() {
        return this.running;
    }
    public void dispose() {
        //assResults.setVisible(false);
        //assResults.dispose();
        motifFinderFrame = null;
    }

    public void stopRunning() {
        this.running = false;
    }

    /**
     * Flag the selected assignment nodes in its corresponding network.
     */
    public void setFlagNetworkNodes(Assignment ass) {
        Set flagged = new HashSet(this.network.getFlaggedNodes());
        if (flagged.size() > 0)
            this.network.setFlaggedNodes(flagged, false);
        this.network.setFlaggedNodes(ass.getNetworkNodes(), true);
    }

    /**
     * Flag the selected assignment nodes in the motif.
     */
    public void setFlagMotifNodes(Assignment ass) {
        Set flagged = new HashSet(this.motif.getFlaggedNodes());
        if (flagged.size() > 0)
            this.motif.setFlaggedNodes(flagged, false);
        this.motif.setFlaggedNodes(ass.getMotifNodes(), true);
    }

    /**
     * Searches all motif nodes against all network nodes.
     *
     * @return The highest scoring assignment found. All other found assignments
     *         found in variable retAss.
     */
    private Assignment getMotifMatch() {
        if (comparer == null)
             return null;

        // looking for a solution contributing with branchfactor per node
        this.bestScore = branchFactor * (double) motif.getNodeCount();
        // bestScore = Double.MIN_VALUE;

        //blackList = new HashSet<Node>();
        Assignment retAss = new Assignment();
        status.setMaximum(network.getNodeCount() * motif.getNodeCount());
        int i = 0;
        List<Node> netNodeList = new Vector<Node>(network.getNodeCount());
        for (Iterator<Node> ni = network.nodesIterator(); ni.hasNext();) {
            netNodeList.add(ni.next());
        }
        Collections.sort(netNodeList, new Comparator<Node>() {
            public int compare(Node o1, Node o2) {
                return Toolkit.getNodeNeighborMap(network).get(o1).size()-
                       Toolkit.getNodeNeighborMap(network).get(o2).size();
            }
        });
        for (Node nnode : netNodeList){
            //System.gc();
            status.setProgressText("");
            if (nnode == null) continue;
            NetNode netNode = new NetNode(network, nnode);
            List<Node> nneig = new Vector<Node>(Toolkit.getNeighbors(netNode));
            status.setProgressText("Trying protein " + nnode.getIdentifier());
            // nneig.removeAll(blackList);
            for (Iterator<Node> mi = motif.nodesIterator(); mi.hasNext();) {
                status.setValue(i++);
                if (!running) return null;
                // motifFinderFrme.setProgress(++iteration * 1000 / niterations);
                Node mnode = mi.next();
                if (mnode == null) continue;
                NetNode motNode = new NetNode(motif, mnode);
                List<Node> mneig = new Vector<Node>(Toolkit.getNeighbors(motNode));

                HashMap<Node, Node> assignments = new HashMap<Node, Node>();
                assignments.put(mnode, nnode);

                double score = comparer.similarity(netNode, motNode);

                Assignment startAss = new Assignment(assignments, score);
                Assignment bestAssignment = match(startAss, mneig, nneig);
                if (bestAssignment == null) continue;
                if (bestAssignment.score > retAss.score)
                    retAss = bestAssignment;
                //if (bestAssignment.score > 0)
                    //assResults.addAssignment(bestAssignment);

            }
            //all possibilities have been tried with nnode. This should now be blacklisted
            //from any other assignment
            if(nneig.size()==1){
                // blackList.add(nnode);
            }

        }
        Toolkit.networkNeigMap.clear();
        assResults.purgeLessScore(bestScore * branchFactor);
        return retAss;
    }

    /**
     * Matching two neighborhoods of an already existing assignment.
     * The best matching is returned.
     *
     * @param ass   The current assignment from a set of motif nodes to a set of network nodes.
     * @param mneig <tt>ass</tt>'s motif neighborhood
     * @param nneig <tt>ass</tt>'s network neighborhood
     * @return The new assignment; the best valid matching of the two neighborhoods.
     */
    private Assignment match(Assignment ass, List<Node> mneig,
                             List<Node> nneig) {

        //remove blacklisted nodes
        //for (Iterator<Node> iter = nneig.iterator(); iter.hasNext();)
         //   if (blackList.contains(iter.next()))  iter.remove();

        // abort if motif's neighborhood is larger than network's
        if (mneig.size() > nneig.size()) {
            assResults.addAssignment(ass);
            return ass;
        }
        // no more nodes to match, return
        if (mneig.size() == 0) {
            // add a full assignment to the assignment list.
            if (ass.score > bestScore) bestScore = ass.score;
            // maybe more searches down a tree gives a perfect assignment
            if (assResults != null && ass.score >= bestScore*branchFactor) {
                assResults.addAssignment(ass);
            }
            return ass;
        }

        // Calculate the max score contributions in branches after this one
        double maxAfter = motif.getNodeCount() - ass.size() - mneig.size();

        // finally, return the best assignment
        Assignment bestAss = new Assignment();

        // creating the new motif boundry for all next recursions
        Vector<Node> motifBoundry = getUnassignedNeighbours(mneig, ass
                .getMotifNodes(), motif);
        motifBoundry.removeAll(mneig);

        // generate and traverse all possible m and n mathes
        List<List<Integer>> matchIndexes = Toolkit.getPermutations(mneig.size(), nneig.size());

        for (int i = 0; i < matchIndexes.size(); i++) {
            List<Integer> mi = matchIndexes.get(i);
            // records this match' score and node-assignment
            double newScore = ass.score;
            Assignment newAssignment = ass.clone();

            // container for the net nodes, to be able to find the unassigned
            // nodes later
            Set<Node> newAssNet = new HashSet<Node>();

            // apply and save assignments/score for this indexset
            boolean allMatched = true;
            for (int j = 0; j < mi.size(); j++) {
                if (!running)
                    return null;
                double free =  ((double) Runtime.getRuntime().freeMemory()) / (double) Runtime.getRuntime().totalMemory();

                if (free < 0.2) {
                    System.runFinalization();
                    System.gc();
                    free = ((double) Runtime.getRuntime().freeMemory()) / (double) Runtime.getRuntime().totalMemory();
                    if (free < 0.2 && moreMemoryRequired) {
                        moreMemoryRequired = true;
                        Thread run = new Thread() {
                            public void run() {
                                JOptionPane.showMessageDialog(null, "Ran out of memory. Some results will be omitted",
                                        "Stack space full",
                                        JOptionPane.ERROR_MESSAGE);
                            }
                        };
                        run.start();
                    }
                    if (free < 0.2) {
                        stopRunning();
                        return null;
                    }
                }
                Node m = mneig.get(j);
                Node n = nneig.get(mi.get(j));
                newAssNet.add(n);


                NetNode nnode = new NetNode(network, n);
                NetNode mnode = new NetNode(motif, m);
                // netnode need connection to an already assigned node to be valid
                if (verifyEdgeAssignments(m, n, ass)) {
                    newScore += comparer.similarity(nnode, mnode);
                    newAssignment.addAssignment(m, n, newScore);
                } else {
                    allMatched = false;
                    /* if the first node doesn't match, it won't match later
                          * either if 1:(0,1) fails on (0), don't try 2:(0,2), 3(0,4)
                          * etc, but goto 10(1,0) instead
                          */
                    if (j == 0) i = i - (i % nneig.size()) + nneig.size();
                    break;
                }
            }

            // abort this recursion as assignment was invalid
            if (!allMatched) continue;
            // BRANCHBOUND

            if (newScore + maxAfter <= bestScore * branchFactor) {
                continue;
            }

            // valid assignments; continue assigments further out in the graph
            //if(newAssignment.score>bestScore) bestScore = newAssignment.score;
            Vector<Node> netBoundry = getUnassignedNeighbours(newAssNet,
                    newAssignment.getNetworkNodes(), network);
            Assignment thisAss = match(newAssignment, motifBoundry, netBoundry);
            // remembering best assignment
            if (thisAss == null) continue;
            if (thisAss.score > bestAss.score) {
                bestAss = thisAss;
            }

        }
        /*if (bestScore < bestAss.score) {
            bestScore = bestAss.score;
        }*/
        return bestAss;
    }

    private boolean verifyEdgeAssignments(Node motifNode, Node netNode,
                                          Assignment assignments) {
        Set<Node> mneighs = motifNeighborMap.get(motifNode);
        Set<Node> nneighs = netNeighborMap.get(netNode);
        Collection<Node> mass = assignments.getMotifNodes();
        mass = new HashSet<Node>(mass);
        mass.retainAll(mneighs);
        for (Node m : mass) {
            Node n = assignments.getAssignedInNetwork(m);
            if (nneighs.contains(n))
                return true;
        }

        return false;
    }

    private Set<Node> getAssignedNeighborsInMotif(Node node, Assignment assigned) {
        Collection<Node> mn = Toolkit.getNeighbors(new NetNode(motif, node));
        Set<Node> assNodes = new HashSet<Node>(assigned.getMotifNodes());
        assNodes.retainAll(mn);
        return assNodes;
    }

    private Vector<Node> getUnassignedNeighbours(Collection<Node> nodes,
                                                 Collection<Node> assigned, CyNetwork network) {
        Set<Node> nodeSet = new HashSet<Node>();
        for (Node node : nodes) {
            Collection<Node> nNeigh = Toolkit.getNeighbors(new NetNode(network, node));
            if (nNeigh != null)
                nodeSet.addAll(nNeigh);
        }
        nodeSet.removeAll(assigned);
        return new Vector<Node>(nodeSet);
    }

    private List<Node> getMostPopNodes(CyNetwork network, int n_nodes) {
        HashMap<Node, Integer> node2nneighbors = new HashMap<Node, Integer>();
        Iterator i = network.nodesIterator();
        while (i.hasNext()) {
            Node node = (Node) i.next();
            int nneighbors = network.getAdjacentEdgeIndicesArray(node
                    .hashCode(), true, true, true).length;
            node2nneighbors.put(node, nneighbors);
        }

        IntegerComparator comp = new IntegerComparator(node2nneighbors);
        Vector<Node> allNodes = new Vector<Node>(node2nneighbors.keySet());
        Collections.sort(allNodes, comp);
        if (n_nodes > allNodes.size())
            return allNodes;
        else
            return allNodes.subList(0, n_nodes - 1);
    }

    private void out(Object o) {
        if (DEBUG)
            System.out.println("MotifFinder: " + o);
    }

    class IntegerComparator implements Comparator<Node> {

        private HashMap<Node, Integer> score;

        public IntegerComparator(HashMap<Node, Integer> score) {
            this.score = score;
        }

        public int compare(Node o1, Node o2) {
            return score.get(o1) - score.get(o2);
        }
    }

    class MotifFinderFrame extends JFrame {
        public MotifFinderFrame() {
            super("MotifFinder");
            Cytoscape.getNetworkSet();
        }
    }

    public static void main(String[] args) {
        List<List<Integer>> matchIndexes = Toolkit.getPermutations(2, 4);
        System.out.println(matchIndexes);

    }

}
