package nemi.protein;

import giny.model.Node;
import nemi.NetNode;
import nemi.compare.ProteinCompare;
import nemi.gui.NeMiFinalView;

import java.text.DecimalFormat;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: kristost
 * Date: 09.mar.2006
 * Time: 17:14:31
 * CSAProteinFinder converts Cytograph node neighborhoods to matrices suitable to the CSA-
 * implementation. compareToAllNeighbours is the main method, takes a node and a network as argument,
 * and return a sorted list of the results. The maximum 100 best results are returned.
 */
public class CSAProteinFinder {
    private static int worst = 5000;
    private static int best = -5000;
    private HashMap<Integer, Double> scoreMap;
    private HashMap<Integer, Integer> trialsMap;

    DecimalFormat f = new DecimalFormat("#.###");
    Object[] allNodes;
    Vector<Node> sourceNeighbors;

    public CSAProteinFinder() {
    }

    /**
     * Compare the source node's neighbors to all other node's neighbors, in the given network,
     * using the provided proteincomparer for node to node scorings.
     */
    public List<CsaMatch> compareToAllNeighbours(NetNode source,
                                                 ProteinCompare comp, NeMiFinalView.StatusInfo status,
                                                 boolean weightedResults) {
        this.allNodes = nemi.Toolkit.getNodeNeighborMap(source.net).keySet().toArray();
        this.sourceNeighbors = new Vector<Node>(nemi.Toolkit.getNeighbors(source));
        scoreMap = new HashMap<Integer, Double>();
        trialsMap = new HashMap<Integer, Integer>();


        List<CsaMatch> matches = new Vector<CsaMatch>();
        int counter = 0;

        status.setMinimum(0);
        status.setMaximum(source.net.getNodeCount());
        status.setStatusInfo("Finding similar proteins to " + source.node.getIdentifier());
        status.setIndeterminate(false);
        status.repaint();
        for (Iterator<Node> iter = (Iterator<Node>) source.net.nodesIterator(); iter.hasNext();) {
            status.setValue(counter++);
            Node n = iter.next();
            if (n.equals(source.node)) continue;
            status.setProgressText("Comparing to " + n.getIdentifier());
            status.repaint();

            NetNode target = new NetNode(source.net, n);

            Vector<Node>targetNeighbors = new Vector<Node>(nemi.Toolkit.getNeighbors(target));

            CsaMatch match = compareNeighbours(source, sourceNeighbors, target, targetNeighbors, comp);

            if (match.score > 0.0) {
                if (weightedResults) {
                    weightResult(match, comp);
                }
                matches.add(match);
            }
            // cleaning the list
            // if (counter > 100 && counter % 100 == 0) cleanList(matches);
        }
        status.setProgressText("");
        // cleanList(matches);
        if (matches.size() == 0) System.out.println("No matches found");
        //nemiView.getProgressBar().setValue(0);
        status.setStatusInfo("Displaying results for " + source.node.getIdentifier());
        return matches;
    }

    /**
     * Weighting a match to how a large set of random target neighbors scores to the match'
     * set of source neighbors. If a random set gives higher score, the the assignment gets
     * lower score and vv.
     *
     * @param match
     * @param comp
     */
    private void weightResult(CSAProteinFinder.CsaMatch match, ProteinCompare comp) {

        Random r = new Random();
        // number of target nodes
        int nt = match.targetNeighs;

        // if (!scoreMap.containsKey(nt)) {
        //double score = 0.0;
        int trials = (int) (1000.0 / (double) nt);
        int randunder = 0;
        for (int i = 0; i < trials; i++) {
            Vector<Node> randRights = new Vector<Node>();
            for (int j = 0; j < nt; j++) {
                randRights.add((Node) allNodes[(int) ((double) allNodes.length * r.nextDouble())]);
            }
            double thisscore = compareNeighbours(match.source, sourceNeighbors, match.target, randRights, comp).score;
            if (thisscore <= match.score) randunder++;
            // score += compareNeighbours(match.source, sourceNeighbors, match.target, randRights, comp).score;
        }
        // score /= (double) trials;
        // scoreMap.put(nt, score);
        // trialsMap.put(nt, trials);
        // }

        // adding weighting info to the match
        match.info += "Weighting: Initial score: " + f.format(match.score) + ", " +
                randunder + " out of " + trials + " random runs was below " + f.format(match.score) + ".\n";

        // changing score
        // match.score = match.score - (scoreMap.get(nt) - match.score);
        match.score = (double)randunder / (double)trials;
        // unexpectedness ratio.
        // match.score = Math.log(match.score/scoreMap.get(nt));

        match.info += "Weighted score: " + f.format(match.score) + "\n";
    }

    private void cleanList(List<CsaMatch> list) {
        if (list.size() > 100) {
            list = list.subList(0, 100);
            Collections.sort(list);
        }
    }

    public CsaMatch compareNeighbours(NetNode source, Vector<Node> neigh1, NetNode target,
                                      Vector<Node> neigh2, ProteinCompare comp) {

        int size = Math.max(neigh1.size(), neigh2.size());
        int[][] weights = new int[size][size];
        for (int y = 0; y < neigh1.size(); y++) {
            NetNode nneigh1 = new NetNode(source.net, neigh1.get(y));
            for (int x = 0; x < neigh2.size(); x++) {
                NetNode nneigh2 = new NetNode(target.net, neigh2.get(x));
                weights[x][y] = (int) (-comp.similarity(nneigh1, nneigh2) * (worst - best) + worst);

            }
            for (int x = neigh2.size(); x < size; x++) {
                weights[x][y] = worst;
            }
        }
        for (int y = neigh1.size(); y < size; y++) {
            for (int x = 0; x < size; x++) {
                weights[x][y] = worst;
            }
        }

        Csa matching = new Csa(weights);
        int[] cost = matching.getMinCost();
        Collection<NodePair> nodePairs = new HashSet<NodePair>();
        double sum = 0.0, interval = -(best - worst);
        boolean firstLargest = (neigh1.size() > neigh2.size());
        Set<Node> unmatched = firstLargest ? new HashSet<Node>(neigh1) : new HashSet<Node>(neigh2);
        for (int i = 0; i < cost.length; i++) {
            if (neigh1.size() > i && neigh2.size() > cost[i]) {
                double weight = weights[i][cost[i]];
                double score = (-best - weight) / interval;
                // if (score > 0) {
                sum += score;
                nodePairs.add(new NodePair(neigh1.get(i), neigh2.get(cost[i]), score));
                if (firstLargest) unmatched.remove(neigh1.get(i));
                else unmatched.remove(neigh2.get(cost[i]));
                // }
            }
        }
        return new CsaMatch(source, neigh1.size(), target, neigh2.size(), nodePairs, sum, unmatched, firstLargest);
    }

    public static void printWeights(int[][] weights) {
        System.out.print(" \t");
        for (int i = 0; i < weights.length; i++) {
            System.out.print(i + "\t");

        }
        System.out.println("");
        int i = 0;
        for (int[] weigthline : weights) {
            System.out.print(i++ + " \t");
            for (int weight : weigthline) {
                System.out.print(weight + "\t");
            }
            System.out.println();
        }
    }

    public class NodePair implements Comparable<NodePair> {
        Node node1, node2;
        double score;
        DecimalFormat form = new DecimalFormat("#.###");

        public NodePair(Node node1, Node node2, double score) {
            this.node1 = node1;
            this.node2 = node2;
            this.score = score;
        }

        public String toString() {
            return node1.getIdentifier() + " match " + node2.getIdentifier() + ", score " + form.format(score);
        }

        public int compareTo(NodePair o) {
            if (score < o.score) return 1;
            else if (score > o.score) return -1;
            else return 0;
        }


    }

    public class CsaMatch implements Comparable<CsaMatch> {
        public NetNode source, target;
        public int sourceNeighs, targetNeighs;
        public Vector<NodePair> matches;
        public Collection<Node> misses;
        public boolean sourcemisses;
        DecimalFormat form = new DecimalFormat("#.###");
        public String info = "";
        public double score = -Double.MAX_VALUE;

        public CsaMatch(NetNode source, int sourceNeighs, NetNode target, int targetNeighs, Collection<NodePair> matches, double score, Collection<Node> misses, boolean sourcemisses) {
            this.source = source;
            this.target = target;
            this.matches = new Vector<NodePair>(matches);
            this.score = score;
            this.sourceNeighs = sourceNeighs;
            this.targetNeighs = targetNeighs;
            this.misses = misses;
            this.sourcemisses = sourcemisses;
            Collections.sort(this.matches);
        }

        public String toString() {
            return target.node.getIdentifier() + " (" + targetNeighs + ")" + " and "
                    + source.node.getIdentifier() + " (" + sourceNeighs + ")";
        }

        public String scoreToString() {
            return form.format(score);
        }


        public String toStringDetailed() {
            StringBuilder sb = new StringBuilder();
            sb.append("Match " + toString());
            sb.append(", score " + form.format(score) + "\n");
            sb.append(info);
            for (NodePair pair : matches) {
                sb.append("-" + pair.toString() + "\n");
            }
            if (misses.size() > 0) {
                if (sourcemisses) sb.append("No mathes for sourcenode");
                else sb.append("No matches for targetnode");
                if (misses.size() > 1) sb.append("s");
                sb.append(":\n");
                for (Node node : misses) {
                    sb.append("-" + node.getIdentifier() + "\n");
                }
            }
            return sb.toString() + "\n";
        }

        public int compareTo(CsaMatch o) {
            if (this.score > o.score) return 1;
            else if (this.score < o.score) return -1;
            else return 0;
        }
    }
}
