package nemi.protein;

import java.util.Stack;

/**
 * Created by IntelliJ IDEA.
 * User: kristoffer.stenersen@gmail.com
 * Date: 29.jan.2006
 * Time: 12:28:46
 *
 * Java implementation of Andrew V. Goldberg's Cost Scaling Algorithm.
 * "An efficient cost scaling algorithm for the assignment problem" by
 * Andrew V. Goldberg and Robert Kennedy, 1995
 *
 */
public final class Csa {
    // the number of nodes
    private final int n;

    private double e;
    private static final double alpha = 10;
    // minor t odo: could be removed and calculated implicitly
    private final double[] xprices;
    private final double[] yprices;
    // cost array
    private final int[][] c;
    // flows from node x into node y
    private final int[] xflows;
    private final int[] yflows;
    // active x nodes (they don't have any flow). stack suggested by article
    private final Stack<Integer> activeNodes;

    public Csa(int[][] weights) {
        c = weights;
        n = c.length;
        e = absMax(c); // the martix' largest value

        xprices = new double[n];
        yprices = new double[n];

        xflows = new int[n];
        yflows = new int[n];

        activeNodes = new Stack<Integer>();
    }

    public int[] getMinCost() {
        if (n==1) return new int[] {0};
        double termVal = 1.0 / (double) n;
        while (e >= termVal) {
            refine();
        }
        return xflows;
    }

    private void refine() {
        // decrease e
        e /= alpha;
        // resets xflows array to -1s
        resetFlow();
        // set min partial reduced cost to every x
        createPseudoflow();
        // loop until f is a flow
        while (activeNodes.size() > 0) {
            doublePush(activeNodes.pop());
        }
    }

    // subroutine for the refine method; updates the x-prices, generates pseudoflows
    private void createPseudoflow() {
        // find the min partial reduced cost c'(a) = c(a) - p(y)
        for (int x = 0; x < xprices.length; x++) {
            double min = Double.MAX_VALUE;
            for (int y = 0; y < c[x].length; y++) {
                double redCost = c[x][y] - yprices[y];
                if (redCost < min) {
                    min = redCost;
                }
            }
            xprices[x] = -min;
        }
    }

    // resets flow arrays
    private void resetFlow() {
        for (int i = 0; i < n; i++) {
            xflows[i] = yflows[i] = -1;
            activeNodes.push(i);
        }
    }

    // find the first admissible arc whose tail is active, enable flow
    private void doublePush(int v) {
        // (v,w) and (v,z) are the smallest and second smallest reduces cost arcs
        int[] smallestRed = this.findTwoSmallestPotentialFrom(v);
        int w = smallestRed[0];
        int z = smallestRed[1];

        // if any, remembering w's old source
        int oldFlowFrom = yflows[w];
        // pushing flow into (v,w)
        xflows[v] = w;
        yflows[w] = v;
        // relabeling v
        xprices[v] = -(c[v][z] - yprices[z]);

        // if w had incoming arc before, remove it
        if (oldFlowFrom >= 0) {
            xflows[oldFlowFrom] = -1;
            // the source now becomes an active node                           
            activeNodes.push(oldFlowFrom);
        }
        // relabel w
        yprices[w] = xprices[v] + c[v][w] - e;

    }

    // finds the two smallest reduced-cost arcs from node x
    private int[] findTwoSmallestPotentialFrom(int x) {
        int ix1 = -1;
        int ix2 = -1;
        double small = Double.MAX_VALUE;
        double smallest = Double.MAX_VALUE;

        for (int y = 0; y < n; y++) {
            double red = c[x][y] - xprices[x] - yprices[y];
            if (red < smallest) {
                small = smallest;
                ix2 = ix1;
                smallest = red;
                ix1 = y;
            } else if (red < small) {
                small = red;
                ix2 = y;
            }
        }
        return new int[] {ix1, ix2};
    }

//    Unused by this version
//    // sets node-price to smalles allowed e-optimal value
//    private void relabel() {
//        // x and y nodes' prices calculated differently
//        for (int x = 0; x < n; x++) {
//            double max = -Double.MAX_VALUE;
//            for (int y = 0; y < n; y++) {
//                double tmax = yprices[y] - (double) c[x][y];
//                if (tmax > max) {
//                    max = tmax;
//                }
//            }
//            xprices[x] = max;
//        }
//        for (int y = 0; y < n; y++) {
//            double max = -Double.MAX_VALUE;
//            for (int x = 0; x < n; x++) {
//                double tmax = xprices[x] - (-(double) c[y][x]) - e;
//                if (max < tmax) {
//                    max = tmax;
//                }
//            }
//            yprices[y] = max;
//        }
//    }
// --Commented out by Inspection STOP (08.06.06 10:33)


    private static int absMax(int[][] fromTable) {
        int max = Integer.MIN_VALUE;
        for (int[] rows : fromTable) {
            for (int val : rows) {
                int aval = Math.abs(val);
                if (aval > max) max = aval;
            }
        }
        return max;
    }



    public static void main(String[] args) {
        /* EXAMPLE USAGE */
        //Csa c = new Csa(new int[][]{{-1,2},{1,3}});
        Csa c = new Csa(new int[][]{{4, 1, 3}, {2, 2, 5}, {7, 6, 5}});
        // Csa c = new Csa(new int[][]{{0, 1, 3,4}, {-1, 2, -3,5}, {3, 2, -2,6},{2,3,40,50}});
        c.getMinCost();
    }
}
