package nemi.protein;

import nemi.Toolkit;

/**
 * The Hungarian Method, solving the maximum bipartite matching problem.
 */
public final class HungarianMethod {


    public static void main(String[] param) {
        double[][] test = {{3, 2, 1}, {2, 5, 4}, {3, 3, 2}};
        Toolkit.printTable(test, "LAP");
        //double[][] test={{1, 2, 3, 4}, {2, 4, 6, 8}, {3, 6, 9, 12}, {4, 8, 12, 16}};

        byte[][] solution = new HungarianMethod(test).munkres();
        Toolkit.printTable(solution, "LAP solution");
    }


    double[][] cost;
    int nCols;
    int nRows;


    public HungarianMethod(double[][] cost) {
        nCols = cost[0].length;
        nRows = cost.length;
        if (nRows > nCols) throw new IllegalArgumentException("Expecting more rows than columns");
        this.cost = cost;
    }


    public byte[][] munkres() {
        munkres m = new munkres();
        return m.solve();
    }


    private static byte NOT_MARKED = 0;
    private static byte STARRED_ZERO = 1;
    private static byte PRIMED_ZERO = 2;


    public class munkres {
        byte[][] mark;
        int[] rowIsCovered;
        int[] colIsCovered;


        public byte[][] solve() {
            mark = new byte[nRows][];
            for (int i = 0; i < nRows; i++) {
                mark[i] = new byte[nCols];
            }

            rowIsCovered = new int[nRows];
            colIsCovered = new int[nCols];

            reduceRows();
            starSomeZeros();
            while (true) {
                if (isDone()) break;
                ;
                pair p = findAndPrimeUncoveredZero();
                new step5().main(p);
            }

            return mark;
        }


        public void reduceRows() {
            double minval;

            for (int i = 0; i < nRows; i++) {
                minval = cost[i][0];
                for (int j = 1; j < nCols; j++) {
                    if (minval > cost[i][j]) {
                        minval = cost[i][j];
                    }
                }

                for (int j = 0; j < nCols; j++) {
                    cost[i][j] -= minval;
                }

            }
        }


        public void starSomeZeros() {
            boolean[] colIsCovered = new boolean[nCols];

            A:
            for (int i = 0; i < nRows; i++) {
                for (int j = 0; j < nCols; j++) {
                    if (!colIsCovered[j] && cost[i][j] == 0) {
                        mark[i][j] = STARRED_ZERO;
                        colIsCovered[j] = true;
                        continue A;
                    }
                }
            }
        }


        public boolean isDone() {
            int count = 0;
            for (int j = 0; j < nCols; j++) {
                for (int i = 0; i < nRows; i++) {
                    if (mark[i][j] == STARRED_ZERO) {
                        colIsCovered[j] = 1;
                        count++;
                        break;
                    }
                }
            }

            if (count >= nRows) return true;
            return false;
        }


        public pair findAndPrimeUncoveredZero() {
            s4:
            while (true) {
                pair p = findUncoveredZero();
                if (p == null) {
                    makeSomeZeros();
                    continue s4;
                }

                mark[p.row][p.col] = PRIMED_ZERO;

                for (int j = 0; j < nCols; j++) {
                    if (mark[p.row][j] == STARRED_ZERO) {
                        rowIsCovered[p.row] = 1;
                        colIsCovered[j] = 0;
                        continue s4;
                    }
                }

                return p;
            }
        }


        public pair findUncoveredZero() {
            for (int i = 0; i < nRows; i++) {
                for (int j = 0; j < nCols; j++) {
                    if (cost[i][j] == 0 && rowIsCovered[i] == 0 && colIsCovered[j] == 0) {
                        return new pair(i, j);
                    }
                }
            }
            return null;
        }

        public void makeSomeZeros() {
            double minval = Float.MAX_VALUE;
            for (int i = 0; i < nRows; i++) {
                if (rowIsCovered[i] == 1) continue;
                for (int j = 0; j < nCols; j++) {
                    if (colIsCovered[j] == 0) {
                        if (minval > cost[i][j]) minval = cost[i][j];
                    }
                }
            }

            for (int i = 0; i < nRows; i++) {
                for (int j = 0; j < nCols; j++) {
                    if (rowIsCovered[i] == 1) cost[i][j] += minval;
                    if (colIsCovered[j] == 0) cost[i][j] -= minval;
                }
            }
        }


        class step5 {
            pair[] path = new pair[nRows * 2];

            public int find_star_in_col(int c) {
                for (int i = 0; i < nRows; i++) {
                    if (mark[i][c] == STARRED_ZERO) {
                        return i;
                    }
                }
                return -1;
            }

            public int find_prime_in_row(int r) {
                for (int j = 0; j < nCols; j++) {
                    if (mark[r][j] == PRIMED_ZERO) {
                        return j;
                    }
                }
                return -1;
            }


            public void clear_covers() {
                for (int i = 0; i < nRows; i++) {
                    rowIsCovered[i] = 0;
                    colIsCovered[i] = 0;
                }
            }

            public void clear_primes() {
                for (int i = 0; i < nRows; i++) {
                    for (int j = 0; j < nCols; j++) {
                        if (mark[i][j] == PRIMED_ZERO) {
                            mark[i][j] = NOT_MARKED;
                        }
                    }
                }
            }

            public void main(pair p) {
                int count = 0;
                path[count] = p;

                while (true) {
                    int r = find_star_in_col(path[count].col);
                    if (r == -1) break;

                    count++;
                    path[count] = new pair(r, path[count - 1].col);

                    int c = find_prime_in_row(path[count].row);

                    count++;
                    path[count] = new pair(path[count - 1].row, c);
                }

                for (int i = 0; i <= count; i++) {
                    if (mark[path[i].row][path[i].col] == STARRED_ZERO) {
                        mark[path[i].row][path[i].col] = NOT_MARKED;
                    } else {
                        mark[path[i].row][path[i].col] = STARRED_ZERO;
                    }
                }

                clear_covers();
                clear_primes();
            }

        }


        public String toString() {
            StringBuffer output = new StringBuffer();

            for (int j = 0; j < nCols; j++) {
                output.append("\t").append(colIsCovered[j]);
            }
            output.append("\n");

            for (int i = 0; i < nRows; i++) {
                output.append(rowIsCovered[i]);
                for (int j = 0; j < nCols; j++) {
                    output.append("\t").append(cost[i][j]);
                    if (mark[i][j] == STARRED_ZERO) {
                        output.append("*");
                    } else if (mark[i][j] == PRIMED_ZERO) {
                        output.append("'");
                    }
                }
                output.append("\n");
            }
            return output.toString();

        }
    }

    private class pair {
        int row;
        int col;

        public pair(int row, int col) {
            this.row = row;
            this.col = col;
        }
    }
}

